from __future__ import annotations

from django_spire.knowledge.collection.tests.factories import create_test_collection
from django_spire.knowledge.entry.models import Entry


def create_test_entry(**kwargs) -> Entry:
    data = {
        'collection': create_test_collection(),
        'name': 'Video Game Cheat Codes',
        'is_deleted': False,
        'is_active': True
    }

    data.update(kwargs)
    return Entry.objects.create(**data)



from __future__ import annotations

from unittest.mock import Mock, patch

from django.test import RequestFactory, override_settings

from django_spire.ai.chat.intelligence.workflows.chat_workflow import chat_workflow
from django_spire.ai.chat.message_intel import DefaultMessageIntel
from django_spire.ai.chat.router import BaseChatRouter
from django_spire.core.tests.test_cases import BaseTestCase


class MockRouter(BaseChatRouter):
    def workflow(self, request, user_input, message_history=None) -> DefaultMessageIntel:
        return DefaultMessageIntel(text='Mock response')


class TestChatWorkflow(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.factory = RequestFactory()
        self.request = self.factory.get('/')
        self.request.user = self.super_user

    @override_settings(
        DJANGO_SPIRE_AI_DEFAULT_CHAT_ROUTER='SPIRE',
        DJANGO_SPIRE_AI_CHAT_ROUTERS={
            'SPIRE': 'django_spire.ai.chat.router.SpireChatRouter'
        }
    )
    def test_workflow_loads_default_router(self) -> None:
        with patch('django_spire.ai.chat.intelligence.workflows.chat_workflow.get_callable_from_module_string_and_validate_arguments') as mock_get_callable:
            mock_router_class = Mock()
            mock_router_instance = Mock()
            mock_router_instance.process.return_value = DefaultMessageIntel(text='Response')
            mock_router_class.return_value = mock_router_instance
            mock_get_callable.return_value = mock_router_class

            result = chat_workflow(
                request=self.request,
                user_input='Hello',
                message_history=None
            )

            mock_get_callable.assert_called_once_with(
                'django_spire.ai.chat.router.SpireChatRouter',
                []
            )

            assert isinstance(result, DefaultMessageIntel)

    @override_settings(
        DJANGO_SPIRE_AI_DEFAULT_CHAT_ROUTER='CUSTOM',
        DJANGO_SPIRE_AI_CHAT_ROUTERS={
            'CUSTOM': 'django_spire.ai.chat.tests.test_router.test_chat_workflow.MockRouter'
        }
    )
    def test_workflow_loads_custom_router(self) -> None:
        result = chat_workflow(
            request=self.request,
            user_input='Hello',
            message_history=None
        )

        assert isinstance(result, DefaultMessageIntel)
        assert result.text == 'Mock response'

    @override_settings(
        DJANGO_SPIRE_AI_DEFAULT_CHAT_ROUTER='NONEXISTENT',
        DJANGO_SPIRE_AI_CHAT_ROUTERS={
            'SPIRE': 'django_spire.ai.chat.router.SpireChatRouter'
        }
    )
    def test_workflow_falls_back_when_router_not_found(self) -> None:
        with patch('django_spire.ai.chat.intelligence.workflows.chat_workflow.get_callable_from_module_string_and_validate_arguments') as mock_get_callable:
            mock_router_class = Mock()
            mock_router_instance = Mock()
            mock_router_instance.process.return_value = DefaultMessageIntel(text='Fallback')
            mock_router_class.return_value = mock_router_instance
            mock_get_callable.return_value = mock_router_class

            result = chat_workflow(
                request=self.request,
                user_input='Hello',
                message_history=None
            )

            assert isinstance(result, DefaultMessageIntel)

    @override_settings(
        DJANGO_SPIRE_AI_DEFAULT_CHAT_ROUTER='TEST',
        DJANGO_SPIRE_AI_CHAT_ROUTERS={
            'TEST': 'django_spire.ai.chat.tests.test_router.test_chat_workflow.MockRouter'
        }
    )
    def test_workflow_passes_message_history(self) -> None:
        from dandy.llm.request.message import MessageHistory

        message_history = MessageHistory()

        with patch('django_spire.ai.chat.tests.test_router.test_chat_workflow.MockRouter.process') as mock_process:
            mock_process.return_value = DefaultMessageIntel(text='Response')

            chat_workflow(
                request=self.request,
                user_input='Hello',
                message_history=message_history
            )

            mock_process.assert_called_once()
            call_kwargs = mock_process.call_args[1]
            assert call_kwargs['message_history'] == message_history



from django.test import TestCase

from django_spire.core.tag.models import Tag
from django_spire.knowledge.collection.models import Collection


class TestTags(TestCase):
    def setUp(self):
        self.collection = Collection.objects.create(
            name='Test Collection', description='Test Collection Description'
        )

        self.tag_set_a = {'apple', 'banana', 'orange'}
        self.tag_set_b = {'grape', 'kiwi', 'pear'}
        self.tag_set_c = {'lemon', 'lime', 'mango'}

        self.mixed_tag_set = {'apple', 'grape', 'lemon'}

        self.dirty_tag_set = {'HaT~', 'four* tacos', '123 hoop!!'}

        self.collection.services.tag.add_tags_from_tag_set(self.tag_set_a)
        self.collection.services.tag.add_tags_from_tag_set(self.tag_set_b)
        self.collection.services.tag.add_tags_from_tag_set(self.tag_set_c)

        self.collection.services.tag.add_tags_from_tag_set(self.mixed_tag_set)

        self.collection.services.tag.add_tags_from_tag_set(self.dirty_tag_set)

    def test_tag_removal(self):
        self.assertIn('apple', self.collection.tag_set)

        self.collection.services.tag.remove_tags_by_tag_set(self.mixed_tag_set)

        self.assertNotIn('lemon', self.collection.tag_set)

        self.assertIn('kiwi', self.collection.tag_set)

    def test_tag_add_and_set(self):
        self.collection.services.tag.set_tags_from_tag_set(self.mixed_tag_set)

        self.assertIn('lemon', self.collection.tag_set)

        self.assertNotIn('banana', self.collection.tag_set)

    def test_tag_count(self):
        self.assertEqual(Tag.objects.count(), 12)
        self.assertEqual(self.collection.tags.count(), 12)

    def test_tag_removal(self):
        self.collection.services.tag.remove_tags_by_tag_set(self.tag_set_a)

        self.assertEqual(self.collection.tags.count(), 9)

    def test_tag_bool_methods(self):
        self.collection.services.tag.remove_tags_by_tag_set(self.tag_set_b)

        self.assertTrue(self.collection.services.tag.has_tags_in_tag_set(self.tag_set_c))
        self.assertFalse(self.collection.services.tag.has_tags_in_tag_set(self.mixed_tag_set))

    def test_tag_matching_methods(self):
        self.assertEqual(len(self.collection.services.tag.get_matching_tags_from_tag_set(self.tag_set_a)), 3)

        self.collection.services.tag.remove_tags_by_tag_set(self.tag_set_a)

        self.assertEqual(len(self.collection.services.tag.get_matching_tags_from_tag_set(self.tag_set_a)), 0)
        self.assertEqual(len(self.collection.services.tag.get_matching_tags_from_tag_set(self.tag_set_b)), 3)
        self.assertEqual(len(self.collection.services.tag.get_matching_tags_from_tag_set(self.mixed_tag_set)), 2)

        self.collection.services.tag.remove_tags_by_tag_set(self.tag_set_b)

        self.assertEqual(len(self.collection.services.tag.get_matching_tags_from_tag_set(self.tag_set_b)), 0)

        self.assertEqual(len(self.collection.services.tag.get_matching_tags_from_tag_set(self.mixed_tag_set)), 1)

    def test_tag_metric_methods(self):
        self.assertAlmostEqual(
            self.collection.services.tag.get_matching_percentage_of_tag_set(self.tag_set_a),
            1.00,
            2
        )

        self.assertAlmostEqual(
            self.collection.services.tag.get_matching_percentage_of_model_tags_from_tag_set(self.tag_set_b),
            0.25,
            2
        )

        self.collection.services.tag.remove_tags_by_tag_set(self.mixed_tag_set)

        self.assertAlmostEqual(
            self.collection.services.tag.get_matching_percentage_of_tag_set(self.tag_set_a),
            0.666,
            2
        )


        self.assertAlmostEqual(
            self.collection.services.tag.get_matching_percentage_of_model_tags_from_tag_set(self.tag_set_b),
            0.222,
            2
        )




# Generated by Django 5.2.3 on 2025-08-04 15:39

import django.db.models.deletion
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('django_spire_auth_user', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Collection',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, editable=False)),
                ('is_deleted', models.BooleanField(default=False, editable=False)),
                ('created_datetime', models.DateTimeField(default=django.utils.timezone.localtime, editable=False)),
                ('order', models.PositiveIntegerField(default=0, editable=False)),
                ('name', models.CharField(max_length=255)),
                ('description', models.TextField()),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', related_query_name='child', to='django_spire_knowledge.collection')),
            ],
            options={
                'verbose_name': 'Collection',
                'verbose_name_plural': 'Collections',
                'db_table': 'django_spire_knowledge_collection',
            },
        ),
        migrations.CreateModel(
            name='Entry',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, editable=False)),
                ('is_deleted', models.BooleanField(default=False, editable=False)),
                ('created_datetime', models.DateTimeField(default=django.utils.timezone.localtime, editable=False)),
                ('order', models.PositiveIntegerField(default=0, editable=False)),
                ('name', models.CharField(max_length=255)),
                ('collection', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entries', related_query_name='entry', to='django_spire_knowledge.collection')),
            ],
            options={
                'verbose_name': 'Entry',
                'verbose_name_plural': 'Entries',
                'db_table': 'django_spire_knowledge_entry',
            },
        ),
        migrations.CreateModel(
            name='EntryVersion',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, editable=False)),
                ('is_deleted', models.BooleanField(default=False, editable=False)),
                ('created_datetime', models.DateTimeField(default=django.utils.timezone.localtime, editable=False)),
                ('published_datetime', models.DateTimeField(blank=True, null=True)),
                ('last_edit_datetime', models.DateTimeField(default=django.utils.timezone.now)),
                ('status', models.CharField(choices=[('draft', 'Draft'), ('published', 'Published'), ('archived', 'Archived')], default='draft', max_length=32)),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='entry_versions', related_query_name='entry_version', to='django_spire_auth_user.authuser')),
                ('entry', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='versions', related_query_name='version', to='django_spire_knowledge.entry')),
            ],
            options={
                'verbose_name': 'Entry Version',
                'verbose_name_plural': 'Entry Versions',
                'db_table': 'django_spire_knowledge_entry_version',
            },
        ),
        migrations.AddField(
            model_name='entry',
            name='current_version',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='current_version', related_query_name='current_version', to='django_spire_knowledge.entryversion'),
        ),
        migrations.CreateModel(
            name='EntryVersionBlock',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_active', models.BooleanField(default=True, editable=False)),
                ('is_deleted', models.BooleanField(default=False, editable=False)),
                ('created_datetime', models.DateTimeField(default=django.utils.timezone.localtime, editable=False)),
                ('order', models.PositiveIntegerField(default=0, editable=False)),
                ('type', models.CharField(choices=[('text', 'Text'), ('heading', 'Heading'), ('sub_heading', 'Sub Heading')], default='text', max_length=32)),
                ('_block_data', models.JSONField()),
                ('_text_data', models.TextField()),
                ('version', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='blocks', related_query_name='block', to='django_spire_knowledge.entryversion')),
            ],
            options={
                'verbose_name': 'Block',
                'verbose_name_plural': 'Blocks',
                'db_table': 'django_spire_knowledge_entry_version_block',
            },
        ),
    ]



from dandy import BaseIntel


class SeedingIntel(BaseIntel):
    items: list[dict]

    def __iter__(self):
        return iter(self.items)


class SourceIntel(BaseIntel):
    file_name: str
    python_source_code: str



from __future__ import annotations

from concurrent.futures import as_completed
from concurrent.futures.thread import ThreadPoolExecutor
from typing import TYPE_CHECKING

from markitdown import MarkItDown

from django_spire.knowledge.entry.version.consts import MARKDOWN_AI_CHUNK_SIZE
from django_spire.knowledge.entry.version.converters.converter import \
    BaseConverter
from django_spire.knowledge.entry.version.block import models
from django_spire.knowledge.entry.version.converters.markdown_converter import \
    MarkdownConverter
from django_spire.knowledge.entry.version.intelligence.bots.markdown_format_llm_bot import \
    MarkdownFormatLlmBot

if TYPE_CHECKING:
    from django_spire.file.models import File


class DocxConverter(BaseConverter):
    """Converts DocX content to a list of EntryVersionBlocks using Markitdown, AI and
    the MarkdownConverter.

    For more info on Markitdown:
    https://github.com/microsoft/markitdown
    """

    def convert_file_to_blocks(self, file: File) -> list[models.EntryVersionBlock]:
        markitdown = MarkItDown()
        markdown_result = markitdown.convert(file.file.url)
        markdown_content = markdown_result.markdown

        markdown_converter = MarkdownConverter(entry_version=self.entry_version)
        return markdown_converter.convert_markdown_to_blocks(
            markdown_content=self.improve_markdown_structure(markdown_content)
        )

    @staticmethod
    def improve_markdown_structure(markdown_content: str) -> str:
        markdown_content_chunks = [
            markdown_content[i: i + MARKDOWN_AI_CHUNK_SIZE]
            for i in range(0, len(markdown_content), MARKDOWN_AI_CHUNK_SIZE)
        ]

        with ThreadPoolExecutor(max_workers=4) as executor:
            futures = []
            for idx, chunk in enumerate(markdown_content_chunks):
                future = executor.submit(
                    MarkdownFormatLlmBot().process,
                    markdown_content=chunk
                )
                future.index = idx
                futures.append(future)

            improved_chunks = {}
            for future in as_completed(futures):
                try:
                    result = future.result(timeout=60)
                    improved_chunks[future.index] = result
                except Exception:
                    improved_chunks[future.index] = markdown_content_chunks[future.index]

        sorted_improved_chunks = sorted(improved_chunks.items(), key=lambda x: x[0])
        return ''.join(chunk[1] for chunk in sorted_improved_chunks)



from unittest import TestCase

from django_spire.core.tag.intelligence.tag_set_bot import TagSetBot

TEST_INPUT = """
    I love reading books about science fiction, fantasy, and adventure, especially ones
    that feature artificial intelligence, machine learning, and natural language processing.
    Some of my favorite authors include Isaac Asimov, J.R.R. Tolkien, and Neil Gaiman.
    I'm also interested in learning more about data science, programming languages like Python
    and Java, and emerging technologies like blockchain and the Internet of Things (IoT).
    Can you recommend some books or resources that align with these interests?
    """


class TestTagIntelligence(TestCase):
    def setUp(self):
        pass

    def test_tag_set_bot(self):
        tag_set = TagSetBot().process(TEST_INPUT)

        self.assertIn('science', tag_set)
        self.assertIn('artificial', tag_set)
        self.assertIn('fantasy', tag_set)

        self.assertNotIn('camping', tag_set)
        self.assertNotIn('art', tag_set)
        self.assertNotIn('hate', tag_set)



from __future__ import annotations

import json

from typing import TYPE_CHECKING

from django.contrib.auth.models import AnonymousUser
from django.template.response import TemplateResponse

from django_spire.notification.app.models import AppNotification

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


def notification_dropdown_template_view(request: WSGIRequest) -> TemplateResponse:
    if isinstance(request.user, AnonymousUser):
        app_notification_list = []

    else:
        app_notification_list = (
            AppNotification.objects.active()
            .is_sent()
            .annotate_is_viewed_by_user(request.user)
            .select_related('notification')
            .distinct()
            .ordered_by_priority_and_sent_datetime()
        )

    body_data = json.loads(request.body.decode('utf-8'))

    return TemplateResponse(
        request,
        context={
            'app_notification_list': app_notification_list,
            'app_notification_list_url': body_data.get('app_notification_list_url'),
        },
        template='django_spire/notification/app/dropdown/notification_dropdown_content.html'
    )



from django.db.models import Model

from dandy import Prompt


def foreign_key_selection_prompt(
    model_class: type[Model],
    related_model_class: type[Model]
) -> Prompt:
    model_name = model_class._meta.verbose_name.title()
    related_model_name = related_model_class._meta.verbose_name.title()
    related_model_plural = related_model_class._meta.verbose_name_plural.title()

    related_model_objects = related_model_class.objects.values_list('id', flat=True)

    return (
        Prompt()
        .heading(f'{related_model_name} Selection')
        .text(f'Each `{model_name}` must be associated with an existing `{related_model_name}`.')
        .text(f'The following "{related_model_plural}" are available to assign:')
        .list(related_model_objects)
    )



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.core.management.commands.spire_startapp_pkg.maps import generate_replacement_map

if TYPE_CHECKING:
    from pathlib import Path

    from django_spire.core.management.commands.spire_startapp_pkg.reporter import Reporter


class TemplateBuilder:
    """
    Builds and displays tree structures for app and template creation.

    This class generates visual representations of the file structure
    that will be created for new apps and their associated templates.
    """

    def __init__(self, reporter: Reporter):
        """
        Initializes the TemplateBuilder with a reporter for output.

        :param reporter: Reporter instance for displaying output to the user.
        """

        self._reporter = reporter

    def build_app_tree_structure(
        self,
        base: Path,
        components: list[str],
        registry: list[str],
        template: Path
    ) -> None:
        """
        Displays a tree structure of the app files that will be created.

        This method shows the user what Python files, directories, and modules
        will be generated for the new Django app before creation.

        :param base: Base directory where the app will be created.
        :param components: List of app path components (e.g., ['app', 'human_resource', 'employee']).
        :param registry: List of already registered apps in the Django project.
        :param template: Path to the app template directory.
        """

        self._reporter.report_tree_structure(
            title='\nThe following app(s) will be created:\n\n',
            base=base,
            components=components,
            registry=registry,
            template=template,
            formatter=self._reporter.format_app_item,
            transformation=self._reporter.transform_app_component,
        )

    def build_html_tree_structure(
        self,
        base: Path,
        components: list[str],
        registry: list[str],
        template: Path
    ) -> None:
        """
        Displays a tree structure of the HTML template files that will be created.

        This method shows the user what HTML templates, cards, forms, and pages
        will be generated for the new Django app before creation.

        :param base: Base directory where templates will be created.
        :param components: List of app path components.
        :param registry: List of already registered apps in the Django project.
        :param template: Path to the HTML template directory.
        """

        replacement = generate_replacement_map(components)

        def html_formatter_with_replacement(item: Path) -> str:
            return self._reporter.format_html_item(item, replacement)

        self._reporter.report_tree_structure(
            title='\nThe following template(s) will be created:\n\n',
            base=base,
            components=components,
            registry=registry,
            template=template,
            formatter=html_formatter_with_replacement,
            transformation=self._reporter.transform_html_component,
        )



from __future__ import annotations

from pydantic import BaseModel


class BaseEditorJsBlockData(BaseModel):
    """
    This class serves as a foundational abstract model for EditorJS tool data objects.

    Note that it does not represent the top level editor block itself,
    but rather the data that is stored within it.
    See https://editorjs.io/getting-started/#tools-installation for a
    list of EditorJS and their associated data models.
    """

    def render_to_text(self) -> str:
        """
        Renders the content to text format.

        This method should be implemented in a subclass to define how the
        content will be rendered into a string. It is meant to be overridden and
        raises a NotImplementedError by default.

        This method is mainly used for providing knowledge base
        content in a digestible format for the AI Chat's LLM connector.

        Returns:
            str: The rendered text output.

        Raises:
            NotImplementedError: If the method is not implemented in a subclass.

        """
        raise NotImplementedError



from __future__ import annotations

from string import Template
from typing import TYPE_CHECKING

from django_spire.core.management.commands.spire_startapp_pkg.maps import generate_replacement_map

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Callable

    from django.core.management.base import BaseCommand


class Reporter:
    """
    Handles user interaction and console output for the app creation command.

    This class manages all console output including status messages, tree
    structures, confirmations, and styled text for the app creation process.
    """

    ICON_FILE = '📄'
    ICON_FOLDER_CLOSED = '📁'
    ICON_FOLDER_OPEN = '📂'
    INDENTATION = '    '

    def __init__(self, command: BaseCommand):
        """
        Initializes the reporter with a Django management command.

        :param command: Django BaseCommand instance for accessing stdout and styling.
        """

        self.command = command
        self.style_error = command.style.ERROR
        self.style_notice = command.style.NOTICE
        self.style_success = command.style.SUCCESS
        self.style_warning = command.style.WARNING

    def format_app_item(self, item: Path) -> str:
        """
        Formats an app file or directory name for display.

        Removes the .template extension from Python files.

        :param item: Path to the item to format.
        :return: Formatted item name.
        """

        return item.name.replace('.py.template', '.py')

    def format_html_item(self, item: Path, replacement: dict[str, str]) -> str:
        """
        Formats an HTML template file or directory name for display.

        Applies variable replacements and removes .template extensions.

        :param item: Path to the item to format.
        :param replacement: Dictionary of placeholder replacements.
        :return: Formatted item name with placeholders replaced.
        """

        if item.is_dir():
            return item.name

        template = Template(item.name)
        filename = template.safe_substitute(replacement)

        if filename.endswith('.template'):
            filename = filename.replace('.template', '')

        return filename

    def prompt_confirmation(self, message: str) -> bool:
        """
        Prompts the user for yes/no confirmation.

        :param message: Confirmation prompt message.
        :return: True if user responds with 'y', False otherwise.
        """

        return input(message).strip().lower() == 'y'

    def report_app_creation_success(self, app: str) -> None:
        """
        Reports successful creation of an app.

        :param app: Name of the app that was created.
        """

        self.write(f'Successfully created app: {app}', self.style_success)

    def report_app_exists(self, app: str, destination: Path) -> None:
        """
        Reports that an app already exists at the destination.

        :param app: Name of the app.
        :param destination: Path where the app already exists.
        """

        self.write(f'The app "{app}" already exists at {destination}', self.style_warning)

    def report_creating_app(self, app: str, destination: Path) -> None:
        """
        Reports that an app is being created.

        :param app: Name of the app being created.
        :param destination: Path where the app will be created.
        """

        self.write(f'Creating app "{app}" at {destination}', self.style_notice)

    def report_creating_templates(self, app: str, destination: Path) -> None:
        """
        Reports that templates are being created for an app.

        :param app: Name of the app.
        :param destination: Path where templates will be created.
        """

        self.write(f'Creating templates for app "{app}" at {destination}', self.style_notice)

    def report_installed_apps_suggestion(self, missing_components: list[str]) -> None:
        """
        Suggests which app to add to INSTALLED_APPS in settings.py.

        :param missing_components: List of missing app components.
        """

        self.write('\nPlease add the following to INSTALLED_APPS in settings.py:', self.style_notice)
        self.write(f'\n {missing_components[-1]}', lambda x: x)

    def report_missing_components(self, missing_components: list[str]) -> None:
        """
        Reports which app components are not yet registered.

        :param missing_components: List of unregistered app component paths.
        """

        self.write('The following are not registered apps:', self.style_warning)
        self.write('\n'.join(f' - {app}' for app in missing_components), lambda x: x)

    def report_templates_creation_success(self, app: str) -> None:
        """
        Reports successful creation of templates for an app.

        :param app: Name of the app whose templates were created.
        """

        self.write(f'Successfully created templates for app: {app}', self.style_success)

    def report_templates_exist(self, app: str, destination: Path) -> None:
        """
        Reports that templates already exist for an app.

        :param app: Name of the app.
        :param destination: Path where templates already exist.
        """

        self.write(f'The templates for app "{app}" already exist at {destination}', self.style_warning)

    def report_tree_structure(
        self,
        title: str,
        base: Path,
        components: list[str],
        registry: list[str],
        template: Path,
        formatter: Callable[[Path], str],
        transformation: Callable[[int, str], str] | None = None,
    ) -> None:
        """
        Displays a tree structure of files and directories that will be created.

        Shows a hierarchical view of the app structure with appropriate icons
        and formatting for files and directories.

        :param title: Title to display above the tree structure.
        :param base: Base directory path.
        :param components: List of app path components.
        :param registry: List of registered apps.
        :param template: Path to template directory.
        :param formatter: Function to format item names for display.
        :param transformation: Optional function to transform component names.
        """

        if transformation is None:
            transformation = self.transform_app_component

        self.command.stdout.write(title)
        current = base

        for i, component in enumerate(components):
            latest = components[: i + 1]
            replacement = generate_replacement_map(latest)

            component = transformation(i, component)
            current = current / component
            app = '.'.join(latest)
            indent = self.INDENTATION * i

            self.command.stdout.write(f'{indent}{self.ICON_FOLDER_OPEN} {component}/')

            if i == len(components) - 1 and app not in registry and template.exists():
                def local_formatter(item: Path, mapping: dict[str, str] = replacement) -> str:
                    base_name = formatter(item)
                    return self._apply_replacement(base_name, mapping)

                self._show_tree_from_template(template, indent + self.INDENTATION, local_formatter)

    def transform_app_component(self, _index: int, component: str) -> str:
        """
        Transforms an app component name (default: no transformation).

        :param _index: Index of the component in the path.
        :param component: Component name to transform.
        :return: Transformed component name.
        """

        return component

    def transform_html_component(self, index: int, component: str) -> str:
        """
        Transforms an HTML component name (replaces first component with 'templates').

        :param index: Index of the component in the path.
        :param component: Component name to transform.
        :return: Transformed component name.
        """

        return 'templates' if index == 0 else component

    def write(self, message: str, style: Callable[[str], str]) -> None:
        """
        Writes a styled message to the console.

        :param message: Message to display.
        :param style: Styling function to apply to the message.
        """

        self.command.stdout.write(style(message))

    def _apply_replacement(self, name: str, replacement: dict[str, str]) -> str:
        """
        Applies placeholder replacements to a name string.

        :param name: String containing placeholders.
        :param replacement: Dictionary mapping placeholders to values.
        :return: String with placeholders replaced.
        """

        for old, new in replacement.items():
            name = name.replace(old, new)
        return name

    def _show_tree_from_template(
        self,
        template: Path,
        indent: str,
        formatter: Callable[[Path], str]
    ) -> None:
        """
        Recursively displays a tree structure from a template directory.

        :param template: Template directory path to display.
        :param indent: Current indentation level.
        :param formatter: Function to format item names.
        """

        ignore = {'__init__.py', '__pycache__'}

        items = sorted(template.iterdir(), key=self._sort_template_items)

        for item in items:
            if item.name in ignore:
                continue

            icon = self.ICON_FOLDER_CLOSED if item.is_dir() else self.ICON_FILE
            self.command.stdout.write(f'{indent}{icon} {formatter(item)}')

            if item.is_dir():
                self._show_tree_from_template(
                    item,
                    indent + self.INDENTATION,
                    formatter
                )

    def _sort_template_items(self, path: Path) -> tuple[bool, str]:
        """
        Sorts template items with directories first, then by name.

        :param path: Path to sort.
        :return: Tuple for sorting (is_file, lowercase_name).
        """

        return (path.is_file(), path.name.lower())



import time
from collections import defaultdict

from django.utils.timezone import now
from twilio.rest.api.v2010.account.message import MessageInstance
from twilio.rest import Client

from django.conf import settings

from django_spire.notification.choices import NotificationStatusChoices
from django_spire.notification.models import Notification
from django_spire.notification.sms.consts import TWILIO_UNSUCCESSFUL_STATUSES
from django_spire.notification.sms.exceptions import TwilioException, \
    TwilioAPIConcurrentException


class BulkTwilioSMSHelper:
    def __init__(self, notifications: list[Notification], client: Client):
        self.notifications = notifications
        self.client = client
        self.notification_segments = defaultdict(int)

    def _find_notification_segments(self):
        for notification in self.notifications:
            message = f'{notification.title}: {notification.body}'
            segments = len(message) // 160 + 1

            self.notification_segments[notification] = segments

    def send_notifications(self):
        self._find_notification_segments()

        sent_segments = 0
        for notification, segments in self.notification_segments.items():
            try:
                TwilioSMSHelper(notification, self.client).send()
                notification.status = NotificationStatusChoices.SENT
                notification.sent_datetime = now()

                sent_segments += segments

                if sent_segments >= settings.TWILIO_SMS_BATCH_SIZE:
                    time.sleep(60)
                    sent_segments = 0

            except Exception as e:
                if isinstance(e, TwilioAPIConcurrentException):
                    notification.status = NotificationStatusChoices.PENDING

                if isinstance(e, TwilioException):
                    notification.status_message = str(e)
                    notification.status = NotificationStatusChoices.ERRORED
                    raise e

                else:
                    notification.status_message = str(e)
                    notification.status = NotificationStatusChoices.FAILED
                    raise e


class TwilioSMSHelper:
    def __init__(self, notification: Notification, client: Client):
        self.to_phone_number = self._format_phone_number(
            notification.sms.to_phone_number
        )
        self.notification = notification
        self.message = f'{notification.title}: {notification.body}'
        self.client = client

    def _attempt_send(self) -> MessageInstance:
        try:
            return self.client.messages.create(
                to=self.to_phone_number,
                from_=settings.TWILIO_PHONE_NUMBER,
                body=self.message,
                media_url=[self.notification.sms.media_url] if self.notification.sms.media_url else [],
            )
        except Exception as e:
            raise TwilioException(f'Twilio Exception: {str(e)}')

    def send(self):
        response = self._attempt_send()
        self._handle_response(response)

    def _handle_response(self, response: MessageInstance):
        if response.error_code == 429:
            raise TwilioAPIConcurrentException(
                'Twilio API concurrent request limit exceeded.'
            )
        if response.status in TWILIO_UNSUCCESSFUL_STATUSES:
            retry_response = self._attempt_send()

            if retry_response.status in TWILIO_UNSUCCESSFUL_STATUSES:
                raise TwilioException(
                    f'Twilio Exception: code={retry_response.error_code}, '
                    f'message={retry_response.error_message}'
                )

    @staticmethod
    def _format_phone_number(phone_number: str) -> str:
        cleaned_number = ''.join(filter(str.isdigit, phone_number))

        if len(cleaned_number) == 10:
            formatted_number = '1' + cleaned_number
        elif len(cleaned_number) == 11 and cleaned_number.startswith('1'):
            formatted_number = cleaned_number
        else:
            raise TwilioException(
                f'Invalid phone number format: {phone_number}.'
            )

        return '+' + formatted_number



from __future__ import annotations


from dandy import BaseIntel, BaseListIntel

from django_spire.knowledge.entry.models import Entry


class EntryIntel(BaseIntel):
    relevant_heading_text: str
    relevant_block_id: int

    def __str__(self):
        return self.relevant_heading_text

    @property
    def collection(self):
        return self.entry.collection

    @property
    def entry(self):
        return Entry.objects.get_by_version_block_id(self.relevant_block_id)


class EntriesIntel(BaseListIntel[EntryIntel]):
    entry_intel_list: list[EntryIntel]



from unittest import TestCase
from typing import get_origin, get_args, Union

from pydantic import BaseModel

from django.core.validators import MinValueValidator, MaxValueValidator
from django.db import models
import enum

from django_spire.core.converters.to_pydantic import DjangoToPydanticFieldConverter, django_to_pydantic_model


def is_optional(typ):
    return get_origin(typ) is Union and type(None) in get_args(typ)


class TestDjangoModelToPydanticModel(TestCase):
    def test_comprehensive_model_conversion(self):
        class ComprehensiveModel(models.Model):
            # Django auto-adds an "id" primary key field.
            char_field_with_choices = models.CharField(
                max_length=10,
                choices=[("A", "Active"), ("I", "Inactive")],
                default="A",
                help_text="Status of the item",
                unique=True,
                null=False
            )
            char_field_no_choices = models.CharField(
                max_length=20,
                default="default",
                help_text="Description",
                unique=False,
                null=True
            )
            int_field = models.IntegerField(
                validators=[MinValueValidator(0), MaxValueValidator(100)],
                default=50,
                help_text="Count",
                unique=False,
                null=True
            )
            decimal_field = models.DecimalField(
                max_digits=5,
                decimal_places=2,
                default=0.0,
                help_text="Price",
                unique=False,
                null=False
            )
            boolean_field = models.BooleanField(
                default=True,
                help_text="Active",
                null=False
            )

            class Meta:
                managed = False

        PydanticModel = django_to_pydantic_model(ComprehensiveModel)
        self.assertEqual(PydanticModel.__name__, "ComprehensiveModel")

    def test_conversion_with_custom_base_class(self):
        class CustomBase(BaseModel):
            extra_field: str = "extra"

        class CustomModel(models.Model):
            char_field = models.CharField(max_length=10, default="custom")
            class Meta:
                managed = False

        PydanticModel = django_to_pydantic_model(CustomModel, base_class=CustomBase)
        instance = PydanticModel(id=1, char_field="custom", extra_field="extra")

        self.assertEqual(instance.char_field, "custom")
        self.assertEqual(instance.extra_field, "extra")

    def test_include_fields_only(self):
        class SimpleModel1(models.Model):
            field1 = models.CharField(max_length=50, default="value1")
            field2 = models.IntegerField(default=10)
            field3 = models.BooleanField(default=True)
            class Meta:
                managed = False

        PydanticModel = django_to_pydantic_model(SimpleModel1, include_fields=["id", "field1", "field3"])
        fields = PydanticModel.__fields__
        self.assertIn("id", fields)
        self.assertIn("field1", fields)
        self.assertIn("field3", fields)
        self.assertNotIn("field2", fields)

    def test_exclude_fields(self):
        class SimpleModel2(models.Model):
            field1 = models.CharField(max_length=50, default="value1")
            field2 = models.IntegerField(default=10)
            field3 = models.BooleanField(default=True)
            class Meta:
                managed = False

        PydanticModel = django_to_pydantic_model(SimpleModel2, exclude_fields=["field2"])
        fields = PydanticModel.__fields__
        self.assertIn("id", fields)
        self.assertIn("field1", fields)
        self.assertIn("field3", fields)
        self.assertNotIn("field2", fields)

    def test_include_and_exclude(self):
        # Define a simple model with multiple fields.
        class SimpleModel3(models.Model):
            field1 = models.CharField(max_length=50, default="value1")
            field2 = models.IntegerField(default=10)
            field3 = models.BooleanField(default=True)
            field4 = models.CharField(max_length=100, default="value4")
            class Meta:
                managed = False

        PydanticModel = django_to_pydantic_model(
            SimpleModel3,
            include_fields=["id", "field1", "field2", "field3"],
            exclude_fields=["field2"]
        )
        fields = PydanticModel.__fields__
        self.assertIn("id", fields)
        self.assertIn("field1", fields)
        self.assertIn("field3", fields)
        self.assertNotIn("field2", fields)
        self.assertNotIn("field4", fields)


class TestDjangoToPydanticFieldConverter(TestCase):
    # def test_nullable_field_is_optional(self):
    #     class Model(models.Model):
    #         char_field = models.CharField(max_length=10, null=True)
    #
    #         class Meta:
    #             managed = False
    #
    #     field = Model._meta.get_field('char_field')
    #     field_type, _ = DjangoToPydanticFieldConverter(field).build_field()
    #
    #     self.assertTrue(is_optional(field_type))

    def test_field_with_no_default_sets_none(self):
        class Model1(models.Model):
            char_field = models.CharField(max_length=10, null=True)

            class Meta:
                managed = False

        field = Model1._meta.get_field('char_field')
        _, field_info = DjangoToPydanticFieldConverter(field).build_field()

        self.assertIsNone(field_info.default)

    # def test_field_with_default(self):
    #     class Model(models.Model):
    #         char_field = models.CharField(max_length=10, default="abc")
    #
    #         class Meta:
    #             managed = False
    #
    #     field = Model._meta.get_field('char_field')
    #     _, field_info = DjangoToPydanticFieldConverter(field).build_field()
    #
    #     self.assertEqual(field_info.default, "abc")

    def test_help_text_becomes_description(self):
        class Model2(models.Model):
            char_field = models.CharField(max_length=10, help_text="hello")

            class Meta:
                managed = False

        field = Model2._meta.get_field('char_field')
        _, field_info = DjangoToPydanticFieldConverter(field).build_field()

        self.assertEqual(field_info.description, "hello")

    def test_enum_created_for_choices(self):
        class Model3(models.Model):
            status = models.CharField(
                max_length=10,
                choices=[("A", "Active"), ("I", "Inactive")],
                default="A"
            )

            class Meta:
                managed = False

        field = Model3._meta.get_field('status')
        field_type, _ = DjangoToPydanticFieldConverter(field).build_field()

        self.assertTrue(issubclass(field_type, enum.Enum))
        self.assertEqual(set(field_type.__members__), {"A", "I"})

    def test_charfield_constraints_applied(self):
        class Model4(models.Model):
            name = models.CharField(max_length=30)

            class Meta:
                managed = False

        field = Model4._meta.get_field('name')
        field_type, _ = DjangoToPydanticFieldConverter(field).build_field()

        self.assertIn("max_length=30", str(field_type))

    def test_integerfield_validators_become_constraints(self):
        # Todo: test needs to be corrected.
        from django.core.validators import MinValueValidator, MaxValueValidator

        class Model5(models.Model):
            count = models.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(10)])

            class Meta:
                managed = False

        field = Model5._meta.get_field('count')
        field_type, field_info = DjangoToPydanticFieldConverter(field).build_field()
        self.assertIn("int", str(field_type))

    def test_decimalfield_constraints_applied(self):
        class Model6(models.Model):
            amount = models.DecimalField(max_digits=5, decimal_places=2)

            class Meta:
                managed = False

        field = Model6._meta.get_field('amount')
        field_type, _ = DjangoToPydanticFieldConverter(field).build_field()

        self.assertIn("max_digits=5", str(field_type))
        self.assertIn("decimal_places=2", str(field_type))



from __future__ import annotations

import json

from http import HTTPStatus
from typing import TYPE_CHECKING

from django.http import JsonResponse
from django.views.decorators.cache import cache_page
from django.views.decorators.http import require_GET, require_POST

from django_spire.theme.enums import ThemeFamily, ThemeMode
from django_spire.theme.models import Theme
from django_spire.theme.utils import get_theme_cookie_name

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


@cache_page(60 * 60 * 24)
@require_GET
def get_config(request: WSGIRequest) -> JsonResponse:
    config = {
        'families': {},
        'default_family': Theme.DEFAULT_FAMILY.value,
        'default_mode': Theme.DEFAULT_MODE.value,
        'separator': Theme.SEPARATOR,
        'cookie_name': get_theme_cookie_name()
    }

    for family in ThemeFamily:
        config['families'][family.value] = {
            'name': Theme.FAMILY_DISPLAY_NAMES.get(family, family.value),
            'modes': [mode.value for mode in ThemeMode]
        }

    return JsonResponse({
        'success': True,
        'data': config
    })


@require_POST
def set_theme(request: WSGIRequest) -> JsonResponse:
    data = json.loads(request.body)
    theme = data.get('theme')

    if not theme:
        return JsonResponse(
            {'error': 'Theme is required', 'success': False},
            status=HTTPStatus.BAD_REQUEST
        )

    try:
        validated = Theme.from_string(theme)
    except ValueError:
        return JsonResponse(
            {'error': f'Invalid theme: {theme}', 'success': False},
            status=HTTPStatus.BAD_REQUEST
        )

    response = JsonResponse({
        'success': True,
        'theme': validated.to_dict()
    })

    response.set_cookie(
        get_theme_cookie_name(),
        validated.value,
        max_age=31536000
    )

    return response



from __future__ import annotations

from django import forms

from django_spire.knowledge.collection.models import Collection


class CollectionForm(forms.ModelForm):
    class Meta:
        model = Collection
        exclude = ['order']



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.management.base import BaseCommand

from django_spire.core.management.commands.spire_startapp_pkg.builder import TemplateBuilder
from django_spire.core.management.commands.spire_startapp_pkg.config import (
    AppConfigFactory,
    PathConfig,
)
from django_spire.core.management.commands.spire_startapp_pkg.filesystem import FileSystem
from django_spire.core.management.commands.spire_startapp_pkg.generator import (
    AppGenerator,
    # TemplateGenerator,
)
from django_spire.core.management.commands.spire_startapp_pkg.processor import (
    TemplateEngine,
    TemplateProcessor,
)
from django_spire.core.management.commands.spire_startapp_pkg.registry import AppRegistry
from django_spire.core.management.commands.spire_startapp_pkg.reporter import Reporter
from django_spire.core.management.commands.spire_startapp_pkg.resolver import PathResolver
from django_spire.core.management.commands.spire_startapp_pkg.user_input import UserInputCollector
from django_spire.core.management.commands.spire_startapp_pkg.validator import AppValidator

if TYPE_CHECKING:
    from typing import Any


class Command(BaseCommand):
    """
    Django management command for creating custom Spire apps.

    This command guides users through an interactive wizard to create a new Django app
    with a pre-configured structure including models, views, forms, services, and templates.
    It handles nested app structures (e.g., app.parent.child) and validates that all
    parent apps are properly registered before creating child apps.
    """

    help = 'Create a custom Spire app.'

    def handle(self, *args: Any, **kwargs: Any) -> None:
        """
        Main entry point for the management command.

        Orchestrates the entire app creation process by initializing helper classes,
        collecting user input through an interactive wizard, validating the app structure,
        generating app files and templates, and providing feedback to the user.

        :param args: Positional arguments (not used).
        :param kwargs: Keyword arguments (not used).
        """

        filesystem = FileSystem()
        path_config = PathConfig.default()
        path_resolver = PathResolver()
        registry = AppRegistry()
        reporter = Reporter(self)

        template_engine = TemplateEngine()
        template_processor = TemplateProcessor(template_engine, filesystem)

        validator = AppValidator(reporter, registry, path_resolver, filesystem)

        user_input_collector = UserInputCollector(reporter, validator)

        config_factory = AppConfigFactory(path_resolver)

        template_builder = TemplateBuilder(reporter)

        app_generator = AppGenerator(filesystem, template_processor, reporter, path_config)
        # template_generator = TemplateGenerator(filesystem, template_processor, reporter, path_config)

        user_inputs = user_input_collector.collect_all_inputs()
        app_path = user_inputs['app_path']

        validator.validate_app_format(app_path)

        config = config_factory.create_config(app_path, user_inputs)
        validator.validate_root_app(config.components)

        reporter.write(
            f'\nChecking app components: {config.components}\n',
            reporter.style_notice
        )

        missing = registry.get_missing_components(config.components)

        if missing:
            reporter.report_missing_components(missing)

            template_builder.build_app_tree_structure(
                path_resolver.get_base_dir(),
                config.components,
                registry.get_installed_apps(),
                path_config.app_template
            )

            # template_builder.build_html_tree_structure(
            #     path_resolver.get_base_dir(),
            #     config.components,
            #     registry.get_installed_apps(),
            #     path_config.html_template
            # )

            if not reporter.prompt_confirmation('\nProceed with app creation? (y/n): '):
                reporter.write('App creation aborted.', reporter.style_error)
                return

            for module in [missing[-1]]:
                module_config = config_factory.create_config(module, user_inputs)
                app_generator.generate(module_config)
                # template_generator.generate(module_config)

            reporter.report_installed_apps_suggestion(missing)
        else:
            reporter.write('All component(s) exist.', reporter.style_success)



from django.db.models import TextChoices


class PersonRoleChoices(TextChoices):
    ADMIN = 'admin', 'Admin'
    HUMAN_RESOURCES = 'human_resources', 'Human Resources'
    SALES = 'sales', 'Sales'
    MANAGER = 'manager', 'Manager'
    MARKETING = 'marketing', 'Marketing'
    TECHNICAL = 'technical', 'Technical'
    IT_SUPPORT = 'it_support', 'IT Support'



from __future__ import annotations

from typing import TYPE_CHECKING

from django.core.cache import cache

if TYPE_CHECKING:
    from typing import Any


class ProgressTracker:
    def __init__(self, key: str, timeout: int = 300) -> None:
        self.key = f'progress_tracker_{key}'
        self.timeout = timeout

    def update(
        self,
        step: str,
        message: str,
        progress: int,
        **kwargs: Any
    ) -> None:
        data = {
            'step': step,
            'message': message,
            'progress': progress,
            **kwargs
        }

        cache.set(self.key, data, timeout=self.timeout)

    def get(self) -> dict[str, Any] | None:
        return cache.get(self.key)

    def clear(self) -> None:
        cache.delete(self.key)

    def error(self, message: str) -> None:
        self.update('error', message, 0)



from __future__ import annotations

from pydantic import BaseModel

from django_spire.knowledge.entry.version.block.data.list.choices import \
    OrderedListCounterType


class ChecklistItemMeta(BaseModel):
    checked: bool = False


class OrderedListItemMeta(BaseModel):
    start: int | None = None
    counterType: OrderedListCounterType | str | None = None



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Locator, Page


class AttributeElement:
    """Playwright component for django_spire/element/attribute_element.html"""

    def __init__(self, page: Page, container_selector: str = 'body') -> None:
        self.container_selector = container_selector
        self.page = page

    @property
    def container(self) -> Locator:
        return self.page.locator(self.container_selector)

    def get_attribute_by_title(self, title: str) -> Locator:
        return self.container.locator(f'.fs-7.text-app-attribute-color:has-text("{title}")').locator('..')

    def get_value_by_title(self, title: str) -> str:
        attribute = self.get_attribute_by_title(title)
        return attribute.locator('.fs-6, a').inner_text()

    def get_value_href_by_title(self, title: str) -> str | None:
        attribute = self.get_attribute_by_title(title)
        link = attribute.locator('a')

        if link.count() > 0:
            return link.get_attribute('href')

        return None

    def has_attribute(self, title: str) -> bool:
        return self.get_attribute_by_title(title).count() > 0

    def is_value_link(self, title: str) -> bool:
        attribute = self.get_attribute_by_title(title)
        return attribute.locator('a').count() > 0


class AttributeList:
    """Helper for pages with multiple attribute elements"""

    def __init__(self, page: Page, container_selector: str = 'body') -> None:
        self.container_selector = container_selector
        self.page = page

    @property
    def container(self) -> Locator:
        return self.page.locator(self.container_selector)

    @property
    def attributes(self) -> Locator:
        return self.container.locator('.fs-7.text-app-attribute-color')

    def get_all_titles(self) -> list[str]:
        count = self.attributes.count()
        return [self.attributes.nth(i).inner_text() for i in range(count)]

    def get_attribute_count(self) -> int:
        return self.attributes.count()

    def get_values_dict(self) -> dict[str, str]:
        result = {}
        attr_element = AttributeElement(self.page, self.container_selector)

        for title in self.get_all_titles():
            result[title] = attr_element.get_value_by_title(title)

        return result



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Locator, Page


class ThemeSelector:
    """Playwright component for django_spire/theme/element/theme_selector.html"""

    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def icon(self) -> Locator:
        return self.page.locator('.bi-sun-fill, .bi-moon-fill').first

    def click(self) -> None:
        self.icon.click()

    def get_current_mode(self) -> str:
        html = self.page.locator('html')
        return html.get_attribute('data-theme') or 'light'

    def get_current_theme_family(self) -> str:
        html = self.page.locator('html')
        return html.get_attribute('data-theme-family') or ''

    def is_dark_mode(self) -> bool:
        return self.get_current_mode() == 'dark'

    def is_light_mode(self) -> bool:
        return self.get_current_mode() == 'light'

    def is_visible(self) -> bool:
        return self.icon.is_visible()

    def toggle(self) -> None:
        self.click()

    def wait_for_theme_change(self, expected_mode: str, timeout: int = 5000) -> None:
        self.page.wait_for_function(
            f'() => document.documentElement.getAttribute("data-theme") === "{expected_mode}"',
            timeout=timeout
        )



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Locator, Page


class SideNavigation:
    """Playwright component for django_spire/navigation/side_navigation.html"""

    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def container(self) -> Locator:
        return self.page.locator('.side-navigation')

    @property
    def links(self) -> Locator:
        return self.container.locator('a.nav-link')

    @property
    def scroll_container(self) -> Locator:
        return self.container.locator('[x-ref="scroll_container"]')

    def click_link(self, text: str) -> None:
        self.get_link_by_text(text).click()

    def get_link_by_text(self, text: str) -> Locator:
        return self.container.locator(f'a.nav-link:has-text("{text}")')

    def get_link_count(self) -> int:
        return self.links.count()

    def get_link_texts(self) -> list[str]:
        return [self.links.nth(i).inner_text() for i in range(self.get_link_count())]

    def has_link(self, text: str) -> bool:
        return self.get_link_by_text(text).count() > 0

    def is_visible(self) -> bool:
        return self.container.is_visible()


class TopNavigation:
    """Playwright component for django_spire/navigation/top_navigation.html"""

    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def container(self) -> Locator:
        return self.page.locator('.top-navigation')

    @property
    def notification_bell(self) -> Locator:
        return self.container.locator('.bi-bell')

    @property
    def theme_selector(self) -> Locator:
        return self.container.locator('.bi-sun-fill, .bi-moon-fill')

    @property
    def title(self) -> Locator:
        return self.container.locator('.fs-5.fw-bold, .fs-md-1.fw-bold').first

    @property
    def user_menu(self) -> Locator:
        return self.container.locator('.bi-person-circle')

    def click_notification_bell(self) -> None:
        self.notification_bell.click()

    def click_theme_selector(self) -> None:
        self.theme_selector.click()

    def click_user_menu(self) -> None:
        self.user_menu.click()

    def get_title_text(self) -> str:
        return self.title.inner_text()

    def is_visible(self) -> bool:
        return self.container.is_visible()


class UserMenu:
    """Playwright component for user dropdown menu in top navigation"""

    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def menu(self) -> Locator:
        return self.page.locator('.dropdown-menu[aria-labelledby="dropdownMenuButton1"]')

    @property
    def trigger(self) -> Locator:
        return self.page.locator('#dropdownMenuButton1')

    def click_admin_panel(self) -> None:
        self.menu.locator('a:has-text("Admin Panel")').click()

    def click_change_password(self) -> None:
        self.menu.locator('a:has-text("Change Password")').click()

    def click_logout(self) -> None:
        self.menu.locator('a:has-text("Logout")').click()

    def click_theme_dashboard(self) -> None:
        self.menu.locator('a:has-text("Theme Dashboard")').click()

    def is_open(self) -> bool:
        return self.menu.is_visible()

    def open(self) -> None:
        if not self.is_open():
            self.trigger.click()



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.contrib.progress.tracker import ProgressTracker

if TYPE_CHECKING:
    from typing import Any


class ProgressTrackingMixin:
    _tracker: ProgressTracker | None = None

    def get_tracker_key(self) -> str:
        raise NotImplementedError

    @property
    def tracker(self) -> ProgressTracker:
        if self._tracker is None:
            key = self.get_tracker_key()
            self._tracker = ProgressTracker(key)

        return self._tracker

    def update_progress(
        self,
        step: str,
        message: str,
        progress: int,
        **kwargs: Any
    ) -> None:
        self.tracker.update(step, message, progress, **kwargs)

    def progress_error(self, message: str) -> None:
        self.tracker.error(message)



from __future__ import annotations

import functools

from typing import Sequence, TYPE_CHECKING

from django.core.exceptions import PermissionDenied
from django.http import HttpResponseRedirect
from django.urls import reverse

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


def permission_required_decorator_function(
        permissions: str | Sequence[str],
        method,
        request: WSGIRequest,
        *args,
        all_required: bool = True,
        **kwargs
):
    if not request.user.is_authenticated:
        return HttpResponseRedirect(reverse('django_spire:auth:admin:login'))

    if isinstance(permissions, str):
        perms = (permissions,)
    else:
        perms = permissions

    if not all_required:
        for perm in perms:
            if request.user.has_perm(perm):
                return method(request, *args, **kwargs)

    if not request.user.has_perms(perms):
        raise PermissionDenied

    return method(request, *args, **kwargs)


def permission_required(
        *permissions: str,
        all_required: bool = True
):
    def decorator(method):
        @functools.wraps(method)
        def wrapper(request: WSGIRequest, *args, **kwargs):
            return permission_required_decorator_function(
                permissions,
                method,
                request,
                *args,
                all_required=all_required,
                **kwargs
            )

        return wrapper

    return decorator



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from django_spire.auth.controller.controller import AppAuthController
from django_spire.conf import settings
from django_spire.consts import __VERSION__
from django_spire.theme.models import Theme
from django_spire.theme.utils import get_theme_cookie_name

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


def django_spire(request: WSGIRequest) -> dict[str, Any]:
    auth_controller_dict = {}

    for app_name in settings.DJANGO_SPIRE_AUTH_CONTROLLERS:
        auth_controller_dict[app_name] = AppAuthController(app_name, request=request)

    return {
        'DJANGO_SPIRE_VERSION': __VERSION__,
        'AuthController': auth_controller_dict
    }


def theme_context(request: WSGIRequest) -> dict[str, Any]:
    default_string = getattr(
        settings,
        'DJANGO_SPIRE_DEFAULT_THEME',
        'default-light'
    )

    default = Theme.from_string(
        default_string,
        default=Theme.get_default()
    )

    name = get_theme_cookie_name()
    cookie = request.COOKIES.get(name, '')
    theme = Theme.from_string(cookie, default=default)

    path = getattr(
        settings,
        'DJANGO_SPIRE_THEME_PATH',
        '/static/django_spire/css/themes/{family}/app-{mode}.css'
    )

    return {
        'DJANGO_SPIRE_DEFAULT_THEME': default.value,
        'DJANGO_SPIRE_THEME_COOKIE_NAME': name,
        'DJANGO_SPIRE_THEME_PATH': path,
        'theme': theme.to_dict(),
    }



from __future__ import annotations

from django.urls import path

from django_spire.file import views


app_name = 'file'

urlpatterns = [
    path('upload/multiple/ajax',
         views.file_multiple_upload_ajax,
         name='upload_multiple_ajax'),

    path('upload/single/ajax',
         views.file_single_upload_ajax,
         name='upload_single_ajax'),
]



from __future__ import annotations

# These imports are for the migrations to work as a single app "django_spire_knowledge"

from django_spire.knowledge.entry import models as entry_models
from django_spire.knowledge.entry.version import models as entry_version_models
from django_spire.knowledge.entry.version.block import models as \
    entry_version_block_models
from django_spire.knowledge.collection import models as collection_models

# Must assert to stop unused import removals.

assert entry_models
assert entry_version_models
assert entry_version_block_models
assert collection_models



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Locator, Page


class NotificationBell:
    """Playwright component for django_spire/notification/app/element/notification_bell.html"""

    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def badge(self) -> Locator:
        return self.bell.locator('.badge, .position-absolute')

    @property
    def bell(self) -> Locator:
        return self.page.locator('.bi-bell').first

    @property
    def dropdown(self) -> Locator:
        return self.page.locator('.notification-dropdown, [x-show*="notification"]')

    def click(self) -> None:
        self.bell.click()

    def get_badge_count(self) -> int:
        if not self.has_badge():
            return 0

        text = self.badge.inner_text()

        if text.isdigit():
            return int(text)

        return 0

    def has_badge(self) -> bool:
        return self.badge.count() > 0 and self.badge.is_visible()

    def has_notifications(self) -> bool:
        return self.has_badge() and self.get_badge_count() > 0

    def is_dropdown_open(self) -> bool:
        return self.dropdown.is_visible()

    def is_visible(self) -> bool:
        return self.bell.is_visible()

    def open_dropdown(self) -> None:
        if not self.is_dropdown_open():
            self.click()

    def close_dropdown(self) -> None:
        if self.is_dropdown_open():
            self.click()



from __future__ import annotations

from unittest.mock import Mock

from django.contrib.auth.models import Permission, User
from django.test import RequestFactory, override_settings

from django_spire.ai.chat.intelligence.decoders.intent_decoder import generate_intent_decoder
from django_spire.ai.chat.message_intel import DefaultMessageIntel
from django_spire.ai.chat.router import BaseChatRouter
from django_spire.core.tests.test_cases import BaseTestCase


class TestRouter(BaseChatRouter):
    def workflow(self, request, user_input, message_history=None) -> DefaultMessageIntel:
        return DefaultMessageIntel(text='Test response')


class TestIntentDecoder(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.factory = RequestFactory()
        self.request = self.factory.get('/')
        self.request.user = self.super_user

    @override_settings(DJANGO_SPIRE_AI_INTENT_CHAT_ROUTERS={})
    def test_decoder_with_no_intents(self) -> None:
        decoder = generate_intent_decoder(
            request=self.request,
            default_callable=lambda **kwargs: DefaultMessageIntel(text='Default')
        )

        assert decoder is not None
        assert len(decoder.mapping) == 1

    @override_settings(
        DJANGO_SPIRE_AI_INTENT_CHAT_ROUTERS={
            'TEST_INTENT': {
                'INTENT_DESCRIPTION': 'Test intent description',
                'CHAT_ROUTER': 'django_spire.ai.chat.tests.test_router.test_intent_decoder.TestRouter',
            }
        }
    )
    def test_decoder_adds_intent_without_permission(self) -> None:
        decoder = generate_intent_decoder(
            request=self.request,
            default_callable=None
        )

        assert len(decoder.mapping) == 1
        assert 'Test intent description' in decoder.mapping

    @override_settings(
        DJANGO_SPIRE_AI_INTENT_CHAT_ROUTERS={
            'TEST_INTENT': {
                'INTENT_DESCRIPTION': 'Test intent with permission',
                'REQUIRED_PERMISSION': 'auth.add_user',
                'CHAT_ROUTER': 'django_spire.ai.chat.tests.test_router.test_intent_decoder.TestRouter',
            }
        }
    )
    def test_decoder_checks_required_permission(self) -> None:
        permission = Permission.objects.get(codename='add_user')
        self.super_user.user_permissions.add(permission)

        decoder = generate_intent_decoder(
            request=self.request,
            default_callable=None
        )

        assert 'Test intent with permission' in decoder.mapping

    @override_settings(
        DJANGO_SPIRE_AI_INTENT_CHAT_ROUTERS={
            'TEST_INTENT': {
                'INTENT_DESCRIPTION': 'Test intent with permission',
                'REQUIRED_PERMISSION': 'auth.add_user',
                'CHAT_ROUTER': 'django_spire.ai.chat.tests.test_router.test_intent_decoder.TestRouter',
            }
        }
    )
    def test_decoder_excludes_intent_without_permission(self) -> None:
        regular_user = User.objects.create_user(username='regular', password='test')

        request = self.factory.get('/')
        request.user = regular_user

        decoder = generate_intent_decoder(
            request=request,
            default_callable=None
        )

        assert 'Test intent with permission' not in decoder.mapping

    @override_settings(
        DJANGO_SPIRE_AI_INTENT_CHAT_ROUTERS={
            'INVALID_ROUTER': {
                'INTENT_DESCRIPTION': 'Invalid router',
                'CHAT_ROUTER': 'non.existent.router.InvalidRouter',
            }
        }
    )
    def test_decoder_handles_import_error(self) -> None:
        decoder = generate_intent_decoder(
            request=self.request,
            default_callable=None
        )

        assert 'Invalid router' not in decoder.mapping

    def test_decoder_adds_default_callable(self) -> None:
        default_callable = Mock()

        decoder = generate_intent_decoder(
            request=self.request,
            default_callable=default_callable
        )

        assert "None of the above choices match the user's intent" in decoder.mapping

    @override_settings(
        DJANGO_SPIRE_AI_INTENT_CHAT_ROUTERS={
            'INTENT_1': {
                'INTENT_DESCRIPTION': 'First intent',
                'CHAT_ROUTER': 'django_spire.ai.chat.tests.test_router.test_intent_decoder.TestRouter',
            },
            'INTENT_2': {
                'INTENT_DESCRIPTION': 'Second intent',
                'CHAT_ROUTER': 'django_spire.ai.chat.tests.test_router.test_intent_decoder.TestRouter',
            }
        }
    )
    def test_decoder_adds_multiple_intents(self) -> None:
        decoder = generate_intent_decoder(
            request=self.request,
            default_callable=None
        )

        assert 'First intent' in decoder.mapping
        assert 'Second intent' in decoder.mapping
        assert len(decoder.mapping) == 2



from django_spire.contrib.progress.mixins import ProgressTrackingMixin
from django_spire.contrib.progress.tracker import ProgressTracker
from django_spire.contrib.progress.views import sse_stream_view


__all__ = [
    'ProgressTracker',
    'ProgressTrackingMixin',
    'sse_stream_view',
]



from __future__ import annotations

from dandy import BaseIntel, Bot, Prompt, recorder_to_html_file


class MarkdownIntel(BaseIntel):
    markdown_content: str


class MarkdownFormatLlmBot(Bot):
    llm_role = 'Markdown Formater'
    llm_task = 'Improve the markdown formatting of the content provided to make more logical sense.'
    llm_guidelines = (
        Prompt()
        .list([
            'Make sure the relevant heading text is from a heading with mark down formatting.'
            'Do not change any of the content only change the formatting as needed.'
        ])
    )
    llm_intel_class = MarkdownIntel

    @recorder_to_html_file('knowledge_markdown_improvement')
    def process(self, markdown_content: str) -> str:
        markdown_intel = self.llm.prompt_to_intel(prompt=markdown_content)

        return markdown_intel.markdown_content



from abc import abstractmethod, ABC
from typing import Generic

from django.db.models import QuerySet

from django_spire.contrib.constructor.django_model_constructor import TypeDjangoModel
from django_spire.contrib.service import BaseDjangoModelService
from django_spire.core.tag import tools
from django_spire.core.tag.models import Tag
from django_spire.core.tag.tools import get_score_percentage_from_tag_set_weighted


class BaseTagService(
    BaseDjangoModelService[TypeDjangoModel],
    ABC,
    Generic[TypeDjangoModel]
):
    obj: TypeDjangoModel

    @abstractmethod
    def process_and_set_tags(self):
        raise NotImplementedError

    def add_tags_from_tag_set(self, tag_set: set[str]):
        self._update_global_tags_from_set(tag_set)

        self.obj.tags.add(*Tag.objects.in_tag_set(tag_set))

    def clear_tags(self):
        self.obj.tags.clear()

    def get_matching_tags_from_tag_set(self, tag_set: set[str]) -> QuerySet:
        return self.obj.tags.in_tag_set(tag_set)

    def get_matching_percentage_of_tag_set(self, tag_set: set[str]) -> float:
        return tools.get_matching_a_percentage_from_tag_sets(tag_set, self.obj.tag_set)

    def get_matching_percentage_of_model_tags_from_tag_set(self, tag_set: set[str]) -> float:
        return tools.get_matching_b_percentage_from_tag_sets(tag_set, self.obj.tag_set)

    def get_score_percentage_from_tag_set_weighted(self, tag_set: set[str]) -> float:
        return get_score_percentage_from_tag_set_weighted(
            tag_set_actual=tag_set,
            tag_set_reference=self.obj.tag_set
        )

    def get_simplified_and_weighted_tag_dict_above_minimum(self, minimum_weight: int = 1) -> dict[str, int]:
        above_minimum_tag_dict = self.obj.simplified_and_weighted_tag_dict

        return {
            tag: weight
            for tag, weight in above_minimum_tag_dict.items()
            if weight >= minimum_weight
        }

    def has_tags_in_tag_set(self, tag_set: set[str]) -> bool:
        return self.obj.tag_set.issuperset(tag_set)

    def remove_tags_by_tag_set(self, tag_set: set[str]):
        tag_objects = Tag.objects.in_tag_set(tag_set)

        self.obj.tags.remove(*tag_objects)

    def set_tags_from_tag_set(self, tag_set: set[str]):
        self._update_global_tags_from_set(tag_set)

        tag_objects = Tag.objects.in_tag_set(tag_set)
        self.obj.tags.set(tag_objects)

    @staticmethod
    def _update_global_tags_from_set(tag_set: set[str]):
        Tag.add_tags([Tag(name=tag) for tag in tag_set])



import base64
import uuid
from io import BytesIO

from PIL import Image

from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt

from django_spire.notification.sms.exceptions import SmsTemporaryMediaException
from django_spire.notification.sms.models import SmsTemporaryMedia


@csrf_exempt
def external_temporary_media_view(request, external_access_key: uuid.UUID) -> HttpResponse:
    try:
        temporary_media = SmsTemporaryMedia.objects.get(external_access_key=external_access_key)
    except SmsTemporaryMedia.DoesNotExist:
        temporary_media = None

    if temporary_media is None or temporary_media.content == '':
        raise SmsTemporaryMediaException("Content for Temporary Media cannot be empty")

    image = Image.open(
        BytesIO(base64.b64decode(temporary_media.content))
    )
    image = image.convert('P', palette=Image.ADAPTIVE, colors=32)

    buffer = BytesIO()
    image.save(buffer, 'PNG')

    return HttpResponse(
        content=buffer.getvalue(),
        content_type=temporary_media.content_type,
    )



from __future__ import annotations

from django import forms

from django_spire.comment import models
from django_spire.comment.widgets import TaggingWidget


class TaggingField(forms.JSONField):
    user_list_func = None

    def __init__(self, user_list, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.widget = TaggingWidget(user_list)


class CommentForm(forms.ModelForm):
    class Meta:
        model = models.Comment
        fields = ['information']



import functools

from django.core.exceptions import PermissionDenied
from django.core.handlers.wsgi import WSGIRequest

from django_spire.auth.permissions.decorators import permission_required_decorator_function
from django_spire.core.utils import get_object_from_module_string


class BaseAuthController:
    def __init__(
            self,
            request: WSGIRequest | None = None
    ):
        self._request = request

    @property
    def request(self):
        if self._request is None:
            raise Exception('AuthController.request is None')

        return self._request

    @request.setter
    def request(self, value: WSGIRequest):
        self._request = value

    def permission_required(
            self,
            *permissions: str,
            all_required: bool = True
    ):

        def decorator(method):
            @functools.wraps(method)
            def wrapper(request: WSGIRequest, *args, **kwargs):
                self.request = request

                uncallable_permissions = []

                for perm in permissions:
                    callable_permission = (
                        getattr(self, perm) if hasattr(self, perm) else perm
                    )
                    if callable(callable_permission):
                        if not all_required and callable_permission():
                            return method(request, *args, **kwargs)
                        elif not callable_permission():
                            raise PermissionDenied

                    else:
                        uncallable_permissions.append(perm)

                return permission_required_decorator_function(
                    uncallable_permissions,
                    method,
                    self.request,
                    *args,
                    all_required=all_required,
                    **kwargs
                )

            return wrapper

        return decorator


class AppAuthController:
    def __new__(
            cls,
            app_name: str,
            request: WSGIRequest | None = None,
            **kwargs
    ):
        from django_spire.conf import settings

        if app_name not in settings.DJANGO_SPIRE_AUTH_CONTROLLERS:
            raise Exception(f'Controller {app_name} not found in settings.AUTH_CONTROLLERS')

        try:
            return get_object_from_module_string(settings.DJANGO_SPIRE_AUTH_CONTROLLERS[app_name])(request)

        except ModuleNotFoundError:
            raise Exception(f'Auth Controller for {app_name} not found')



from django.urls import path

from django_spire.knowledge.entry.version.views import redirect_views


app_name = 'redirect'

urlpatterns = [
    path('<int:pk>/publish/', redirect_views.publish_view, name='publish'),
]



from __future__ import annotations

from typing import Any

from django.apps import apps
from django.contrib.auth.models import User, Group

from django_spire.auth.permissions.permissions import ModelPermission, UserPermissionHelper, GroupPermissions


def generate_model_permissions() -> list[ModelPermission]:
    model_permissions = []

    for app_config in apps.get_app_configs():
        if hasattr(app_config, 'MODEL_PERMISSIONS'):
            for model_permission in app_config.MODEL_PERMISSIONS:
                model_permissions.append(
                    ModelPermission(**model_permission)
                )

    return model_permissions

def generate_model_key_permission_map() -> dict[str, ModelPermission]:
    return {
        model_permission.name.lower(): model_permission
        for model_permission in generate_model_permissions()
    }

def generate_user_perm_data(user: User) -> list[dict]:
    perm_data = []

    for permission in generate_model_permissions():

        user_permissions = UserPermissionHelper(user, permission)

        perm_data.append({
            'app_name': permission.name.capitalize(),
            'level_verbose': user_permissions.perm_level_verbose()
        })

    return perm_data


def generate_group_perm_data(
        group: Group,
        with_special_role: bool = False
) -> list[dict]:

    perm_data = []

    for model_permission in generate_model_permissions():
        group_permissions = GroupPermissions(group=group, model_permission=model_permission)

        perm_information_dic = {
            'app_name': model_permission.name.capitalize(),
            'verbose_app_name': model_permission.verbose_name,
            'level_verbose': group_permissions.perm_level_verbose()
        }

        if with_special_role:
            perm_information_dic['special_role_data'] = generate_special_role_data(group_permissions)

        perm_data.append(perm_information_dic)

    return perm_data


def generate_special_role_data(
        group_permissions: GroupPermissions
) -> list[dict[str, Any]]:
    special_role_data = []

    model_permissions = group_permissions.model_permissions

    for special_role in model_permissions.special_role_list():
        special_role_data.append({
            'name': special_role.name,
            'codename': special_role.codename,
            'has_access': group_permissions.has_special_role(special_role.codename)
        })

    return special_role_data



from __future__ import annotations

from typing import TypedDict


class BreadcrumbDict(TypedDict):
    name: str
    href: str | None


class BreadcrumbItem:
    def __init__(self, name: str, href: str | None = None):
        self.name = name
        self.href = href

    def to_dict(self) -> BreadcrumbDict:
        return {'name': self.name, 'href': self.href}


class Breadcrumbs:
    def __init__(self, branch_slug: str | None = None):
        self.data: list[BreadcrumbItem] = []
        self.index: int = 0

    def __add__(self, other):
        self.data += other.data
        return self

    def __iter__(self):
        self.index = 0
        return self

    def __str__(self) -> str:
        return str(self.data)

    def __len__(self) -> int:
        return len(self.data)

    def __next__(self) -> BreadcrumbDict:
        if self.index >= len(self.data):
            raise StopIteration

        breadcrumb_item = self.data[self.index]
        self.index += 1
        return breadcrumb_item.to_dict()

    def add_base_breadcrumb(self, model) -> None:
        if hasattr(model, 'base_breadcrumb'):
            self += model.base_breadcrumb()

    def add_breadcrumb(self, name: str, href: str | None = None) -> None:
        breadcrumb_item = BreadcrumbItem(name=name, href=href)
        self.data.append(breadcrumb_item)

    def add_obj_breadcrumbs(self, obj) -> None:
        # Expects a breadcrumb object to be returned from the object
        if hasattr(obj, 'breadcrumbs'):
            object_breadcrumbs = obj.breadcrumbs()
            self.data += object_breadcrumbs.data

    def add_form_breadcrumbs(self, obj) -> None:
        self.add_obj_breadcrumbs(obj)

        if obj.pk is None:
            self.add_breadcrumb(name=obj._meta.model._meta.verbose_name)
            self.add_breadcrumb(name='Create')
        else:
            self.add_breadcrumb(name='Edit')

    def remove(self, index: int):
        del self.data[index]
        return self

    def reverse(self):
        self.data.reverse()
        return self



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Locator


class Accordion:
    """Playwright component for django_spire/accordion/accordion.html"""

    def __init__(self, parent_locator: Locator) -> None:
        self.parent = parent_locator

    @property
    def content(self) -> Locator:
        return self.parent.locator('[x-show="show_accordion"]')

    @property
    def toggle(self) -> Locator:
        return self.parent.locator('[\\@click*="toggle"]').first

    def close(self) -> None:
        if self.is_open():
            self.toggle.click()

    def is_open(self) -> bool:
        return self.content.is_visible()

    def open(self) -> None:
        if not self.is_open():
            self.toggle.click()


class NavAccordion(Accordion):
    """Playwright component for django_spire/navigation/accordion/nav_accordion.html"""

    @property
    def chevron(self) -> Locator:
        return self.parent.locator('.bi-chevron-right, .bi-chevron-down')

    @property
    def icon(self) -> Locator:
        return self.parent.locator('i.fs-6').first

    @property
    def title(self) -> Locator:
        return self.parent.locator('span.h6')

    def get_title_text(self) -> str:
        return self.title.inner_text()

    def is_expanded(self) -> bool:
        chevron_classes = self.chevron.get_attribute('class') or ''
        return 'bi-chevron-down' in chevron_classes



from __future__ import annotations

import django_glue as dg

from typing import TYPE_CHECKING

from django.http import HttpResponseRedirect
from django.urls import reverse

from django_spire.auth.controller.controller import AppAuthController
from django_spire.auth.group.models import AuthGroup
from django_spire.contrib.form.utils import show_form_errors
from django_spire.contrib.generic_views import portal_views
from django_spire.core.shortcuts import get_object_or_null_obj
from django_spire.knowledge.collection.models import Collection, CollectionGroup
from django_spire.knowledge.collection.forms import CollectionForm

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest
    from django.template.response import TemplateResponse


@AppAuthController('knowledge').permission_required('can_add')
def form_view(
        request: WSGIRequest,
        pk: int = 0,
        parent_pk: int = None,
) -> TemplateResponse | HttpResponseRedirect:
    collection = get_object_or_null_obj(Collection, pk=pk)

    dg.glue_model_object(request, unique_name='collection', model_object=collection)

    dg.glue_query_set(
        request,
        unique_name='collections',
        target=(
            Collection.objects
            .active()
            .request_user_has_access(request)
            .exclude(pk=collection.pk)
            .exclude_children(collection_id=pk)
        ),
        fields=['name']
    )
    dg.glue_query_set(
        request,
        unique_name='group_query_set',
        target=AuthGroup.objects.all(),
        fields=['name']
    )

    if request.method == 'POST':
        form = CollectionForm(request.POST, instance=collection)

        if form.is_valid():
            collection, _ = collection.services.save_model_obj(**form.cleaned_data)

            _ = CollectionGroup.services.factory.replace_groups(
                request=request,
                group_pks=dict(request.POST).get('groups'),
                collection=collection,
            )

            collection.services.tag.process_and_set_tags()

            if collection.parent_id:
                return_url = reverse(
                    'django_spire:knowledge:collection:page:top_level',
                    kwargs={'pk': collection.services.tool.get_root_collection_pk()}
                )
            else:
                return_url = reverse('django_spire:knowledge:page:home')

            return HttpResponseRedirect(
                return_url
            )

        show_form_errors(request, form)
    else:
        form = CollectionForm(instance=collection)

    return portal_views.form_view(
        request,
        form=form,
        obj=collection,
        context_data={
            'collection': collection,
            'collection_parent_pk': parent_pk,
            'group_ids': list(
                collection.groups.all().values_list('auth_group_id', flat=True)
            ) if collection.id else [],
        },
        template='django_spire/knowledge/collection/page/form_page.html'
    )



from __future__ import annotations

import pytest

from typing import TYPE_CHECKING

from django.test import RequestFactory

from django_spire.ai.chat.message_intel import BaseMessageIntel, DefaultMessageIntel
from django_spire.ai.chat.router import BaseChatRouter
from django_spire.core.tests.test_cases import BaseTestCase

if TYPE_CHECKING:
    from dandy.llm.request.message import MessageHistory
    from django.core.handlers.wsgi import WSGIRequest


class TestBaseChatRouter(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.factory = RequestFactory()
        self.request = self.factory.get('/')
        self.request.user = self.super_user

    def test_workflow_is_abstract_method(self) -> None:
        with pytest.raises(TypeError):
            BaseChatRouter()

    def test_process_calls_workflow(self) -> None:
        class TestRouter(BaseChatRouter):
            workflow_called = False

            def workflow(
                self,
                request: WSGIRequest,
                user_input: str,
                message_history: MessageHistory | None = None
            ) -> BaseMessageIntel:
                self.workflow_called = True
                return DefaultMessageIntel(text='Test response')

        router = TestRouter()
        result = router.process(
            request=self.request,
            user_input='Hello',
            message_history=None
        )

        assert router.workflow_called
        assert isinstance(result, DefaultMessageIntel)
        assert result.text == 'Test response'

    def test_process_validates_workflow_return_type(self) -> None:
        class InvalidRouter(BaseChatRouter):
            def workflow(self, request, user_input, message_history=None) -> str:
                return 'Invalid return type'

        router = InvalidRouter()

        with pytest.raises(TypeError) as cm:
            router.process(
                request=self.request,
                user_input='Hello',
                message_history=None
            )

        assert 'BaseMessageIntel' in str(cm.value)

    def test_process_handles_none_return(self) -> None:
        class NoneRouter(BaseChatRouter):
            def workflow(self, request, user_input, message_history=None) -> None:
                return None

        router = NoneRouter()
        result = router.process(
            request=self.request,
            user_input='Hello',
            message_history=None
        )

        assert isinstance(result, DefaultMessageIntel)
        assert result.text == 'I apologize, but I was unable to process your request.'

    def test_process_accepts_all_parameters(self) -> None:
        from dandy.llm.request.message import MessageHistory

        class ParamTestRouter(BaseChatRouter):
            received_params = {}

            def workflow(self, request, user_input, message_history=None) -> DefaultMessageIntel:
                self.received_params = {
                    'request': request,
                    'user_input': user_input,
                    'message_history': message_history
                }
                return DefaultMessageIntel(text='Success')

        router = ParamTestRouter()
        message_history = MessageHistory()

        router.process(
            request=self.request,
            user_input='Test input',
            message_history=message_history
        )

        assert router.received_params['request'] == self.request
        assert router.received_params['user_input'] == 'Test input'
        assert router.received_params['message_history'] == message_history



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.conf import settings
from django_spire.core.utils import get_callable_from_module_string_and_validate_arguments

if TYPE_CHECKING:
    from dandy.llm.request.message import MessageHistory
    from django.core.handlers.wsgi import WSGIRequest

    from django_spire.ai.chat.message_intel import BaseMessageIntel


def chat_workflow(
    request: WSGIRequest,
    user_input: str,
    message_history: MessageHistory | None = None
) -> BaseMessageIntel:
    router_key = getattr(settings, 'DJANGO_SPIRE_AI_DEFAULT_CHAT_ROUTER', 'SPIRE')

    chat_routers = getattr(settings, 'DJANGO_SPIRE_AI_CHAT_ROUTERS', {
        'SPIRE': 'django_spire.ai.chat.router.SpireChatRouter'
    })

    router_path = chat_routers.get(router_key)

    if not router_path:
        router_path = 'django_spire.ai.chat.router.SpireChatRouter'

    router_class = get_callable_from_module_string_and_validate_arguments(
        router_path,
        []
    )

    router_instance = router_class()

    return router_instance.process(
        request=request,
        user_input=user_input,
        message_history=message_history,
    )



from django.urls import path

from django_spire.knowledge.collection.views import page_views


app_name = 'page'

urlpatterns = [
    path('<int:pk>/', page_views.top_level_collection_view, name='top_level'),
    path('<int:pk>/delete/', page_views.delete_view, name='delete'),
]



from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.http import HttpResponseRedirect
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.urls import reverse

from django_spire.comment import models
from django_spire.comment.forms import CommentForm
from django_spire.core.redirect import safe_redirect_url
from django_spire.core.shortcuts import get_object_or_null_obj, model_object_from_app_label
from django_spire.contrib.form.utils import show_form_errors
from django_spire.auth.group.utils import has_app_permission_or_404
from django_spire.contrib.generic_views import dispatch_modal_delete_form_content

import django_glue as dg

if TYPE_CHECKING:
    from django.core.handlers.wsgi import WSGIRequest


@login_required()
def comment_modal_form_content(
    request: WSGIRequest,
    comment_pk: int,
    obj_pk: int,
    app_label: str,
    model_name: str
) -> TemplateResponse:
    has_app_permission_or_404(request.user, app_label, model_name, 'change')

    if comment_pk == 0:
        comment = get_object_or_null_obj(models.Comment, pk=comment_pk)
    else:
        comment = get_object_or_404(models.Comment, pk=comment_pk, user__id=request.user.pk)

    dg.glue_model_object(request, 'comment', comment)

    context_data = {
        'app_label': app_label,
        'model_name': model_name,
        'comment': comment,
        'obj_pk': obj_pk,
    }

    return TemplateResponse(
        request,
        context=context_data,
        template='django_spire/comment/form/comment_form.html',
    )


@login_required()
def comment_form_view(
    request: WSGIRequest,
    comment_pk: int,
    obj_pk: int,
    app_label: str,
    model_name: str
) -> HttpResponseRedirect:
    has_app_permission_or_404(request.user, app_label, model_name, 'change')

    if comment_pk == 0:
        comment = get_object_or_null_obj(models.Comment, pk=comment_pk)
    else:
        comment = get_object_or_404(models.Comment, pk=comment_pk, user__id=request.user.pk)

    obj = model_object_from_app_label(app_label, model_name, obj_pk)

    if not hasattr(obj, 'add_comment'):
        message = f'Object {obj} does not have the comment model mixin.'
        raise Exception(message)

    if request.method == 'POST':
        form = CommentForm(request.POST, instance=comment)
        if form.is_valid():
            # Todo: Create comment factory.
            if comment_pk == 0:
                obj.add_comment(
                    user=request.user,
                    information=form.cleaned_data['information']
                )
            else:
                comment.information = form.cleaned_data['information']
                comment.is_edited = True
                comment.save()
        else:
            show_form_errors(request, form)

    return HttpResponseRedirect(safe_redirect_url(request))


@login_required()
def comment_modal_delete_form_view(
    request: WSGIRequest,
    comment_pk: int,
    obj_pk: int,
    app_label: str,
    model_name: str
) -> HttpResponseRedirect | TemplateResponse:
    has_app_permission_or_404(request.user, app_label, model_name, 'change')

    comment = get_object_or_404(models.Comment, pk=comment_pk)
    obj = model_object_from_app_label(app_label, model_name, obj_pk)
    return_url = safe_redirect_url(request)

    if comment.user != request.user:
        messages.warning(request, 'You can only delete your comments.')
        return HttpResponseRedirect(return_url)

    form_action = reverse('comment:delete_form', kwargs={
        'comment_pk': comment_pk,
        'obj_pk': obj_pk,
        'app_label': app_label,
        'model_name': model_name
    })

    def add_activity() -> None:
        obj.add_activity(
            user=request.user,
            verb='deleted',
            device=request.device,
            information=f'{request.user.get_full_name()} deleted a comment on "{obj}".'
        )

    return dispatch_modal_delete_form_content(
        request,
        obj=comment,
        form_action=form_action,
        activity_func=add_activity,
        return_url=return_url,
    )



from __future__ import annotations

from django.apps import AppConfig
from django_spire.utils import check_required_apps


class AiChatConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'django_spire.ai.chat'
    label = 'django_spire_ai_chat'
    MODEL_PERMISSIONS = (
        {
            'name': 'ai_chat',
            'verbose_name': 'AI Chat',
            'model_class_path': 'django_spire.ai.chat.models.Chat',
            'is_proxy_model': False,
        },
    )

    REQUIRED_APPS = ('django_spire_ai', 'django_spire_ai_context')

    def ready(self) -> None:
        check_required_apps(self.label)



from django.contrib.auth.models import Group
from django_spire.auth.user.models import AuthUser


def add_user_to_all_user_group(user: AuthUser):
    try:
        all_user_group = Group.objects.get(name='All Users')
    except Group.DoesNotExist:
        all_user_group = Group.objects.create(name='All Users')

    all_user_group.user_set.add(user)



from unittest.mock import patch

from django.urls import reverse

from django_spire.ai.sms.models import SmsConversation
from django_spire.core.tests.test_cases import BaseTestCase


class SmsWebhookTests(BaseTestCase):
    def setUp(self):
        super().setUp()

        self.webhook_url = reverse('django_spire:ai:sms:webhook')

    @patch('twilio.request_validator.RequestValidator.validate')
    def test_webhook_receives_message(self, mock_validate):
        mock_validate.return_value = True

        response = self.client.post(
            self.webhook_url,
            {
                'From': '+15551234567',
                'Body': 'Hello',
                'MessageSid': 'SM123456789'
            }
        )

        self.assertEqual(response.status_code, 200)

        conversation = SmsConversation.objects.get(phone_number='+15551234567')
        self.assertEqual(conversation.messages.count(), 2)  # Inbound and outbound

        inbound_message = conversation.messages.filter(is_inbound=True).first()
        self.assertEqual(inbound_message.body, 'Hello')
        self.assertEqual(inbound_message.twilio_sid, 'SM123456789')

        outbound_message = conversation.messages.filter(is_inbound=False).first()
        self.assertTrue(len(outbound_message.body) > 0)



from django.urls import path

from django_spire.knowledge.entry.views import json_views


app_name = 'json'

urlpatterns = [
    path('reorder/', json_views.reorder_view, name='reorder'),
    path('update/files/', json_views.update_files_view, name='update_files'),
]



# Generated by Django 5.2.1 on 2025-05-15 20:20

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('django_spire_notification', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='SmsNotification',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('to_phone_number', models.CharField(blank=True, max_length=11)),
                ('notification', models.OneToOneField(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='sms', related_query_name='sms', to='django_spire_notification.notification')),
            ],
            options={
                'verbose_name': 'SMS Notification',
                'verbose_name_plural': 'SMS Notifications',
                'db_table': 'django_spire_notification_sms',
            },
        ),
    ]



from django_spire.testing.playwright.components.accordion import Accordion, NavAccordion
from django_spire.testing.playwright.components.attribute_element import AttributeElement, AttributeList
from django_spire.testing.playwright.components.base_session_filter_form import FilterForm
from django_spire.testing.playwright.components.breadcrumb_element import Breadcrumb
from django_spire.testing.playwright.components.card import Card, FormCard, InfiniteScrollCard, TitleCard
from django_spire.testing.playwright.components.dropdown import Dropdown, EllipsisDropdown, EllipsisModalDropdown, EllipsisTableDropdown
from django_spire.testing.playwright.components.infinite_scroll import InfiniteScroll, InfiniteScrollList, InfiniteScrollTable
from django_spire.testing.playwright.components.lazy_tab import LazyTab
from django_spire.testing.playwright.components.modal import DeleteModal, FormModal, Modal, TitleModal
from django_spire.testing.playwright.components.navigation import SideNavigation, TopNavigation, UserMenu
from django_spire.testing.playwright.components.notification_bell import NotificationBell
from django_spire.testing.playwright.components.theme_selector import ThemeSelector
from django_spire.testing.playwright.components.toast import Toast


__all__ = [
    'Accordion',
    'AttributeElement',
    'AttributeList',
    'Breadcrumb',
    'Card',
    'DeleteModal',
    'Dropdown',
    'EllipsisDropdown',
    'EllipsisModalDropdown',
    'EllipsisTableDropdown',
    'FilterForm',
    'FormCard',
    'FormModal',
    'InfiniteScroll',
    'InfiniteScrollCard',
    'InfiniteScrollList',
    'InfiniteScrollTable',
    'LazyTab',
    'Modal',
    'NavAccordion',
    'NotificationBell',
    'SideNavigation',
    'ThemeSelector',
    'TitleCard',
    'TitleModal',
    'Toast',
    'TopNavigation',
    'UserMenu',
]



from typing import Sequence, Self

from django.db import models
from django.utils.text import slugify

from django_spire.core.tag.querysets import TagQuerySet


class Tag(models.Model):
    name = models.SlugField(
        max_length=128,
        unique=True
    )

    objects = TagQuerySet.as_manager()

    def __init__(self, *args, **kwargs):
        name = kwargs.get('name')
        if name:
            kwargs['name'] = slugify(name)

        super().__init__(*args, **kwargs)

    def __str__(self):
        return self.name

    @classmethod
    def add_tags(cls, tags: Sequence[Self]):
        cls.objects.bulk_create(
            tags,
            ignore_conflicts=True
        )

    class Meta:
        db_table = 'django_spire_core_tag'
        verbose_name = 'Tag'
        verbose_name_plural = 'Tags'



from __future__ import annotations

from typing import TYPE_CHECKING

from dandy import Bot, Prompt

from django_spire.knowledge.intelligence.intel.answer_intel import AnswerIntel

if TYPE_CHECKING:
    from django_spire.knowledge.entry.models import Entry


class KnowledgeAnswerBot(Bot):
    llm_role = 'Knowledge Entry Search Assistant'
    llm_task = 'Read through the knowledge and answer the users request.'
    llm_guidelines = (
        Prompt()
        .list([
            'Make sure the answer is relevant and reflects knowledge entries.',
            'Do not make up information use the provided knowledge entries as a source of truth.',
            'Use line breaks to separate sections of the answer and use 2 if you need to separate the section from the previous.'
        ])
    )
    llm_intel_class = AnswerIntel

    def process(self, user_input: str, entries: list[Entry]) -> AnswerIntel:

        entry_prompt = Prompt()
        entry_prompt.sub_heading('User Request')
        entry_prompt.line_break()
        entry_prompt.text(f'{user_input}')
        entry_prompt.line_break()
        entry_prompt.sub_heading('Knowledge')
        entry_prompt.line_break()

        for entry in entries:
            for version_block in entry.current_version.blocks.all():
                if version_block.render_to_text() != '\n':
                    entry_prompt.text(f'{version_block.render_to_text()}')

        return self.llm.prompt_to_intel(
            prompt=entry_prompt,
        )




from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.contrib.ordering.exceptions import OrderingMixinException

if TYPE_CHECKING:
    from django.db.models import Model, QuerySet


class OrderingMixinValidator:
    def __init__(
        self,
        destination_objects: QuerySet[Model],
        position: int,
        obj: Model,
        origin_objects: QuerySet[Model],
    ):
        self._destination_objects = destination_objects
        self._position = position
        self._origin_objects = origin_objects
        self._obj = obj

        self._errors: list[OrderingMixinException] = []

    @property
    def errors(self) -> list[OrderingMixinException]:
        """Returns list of OrderingMixinException errors."""
        return self._errors

    def validate(self) -> bool:
        """
        Validates that the destination and origin and insertion objects and position are valid.
        Returns tuple of validity and applicable error messages.
        """

        self._validate_position()
        return not self._errors

    def _validate_position(self):
        """Ensure position is valid."""
        if self._position < 0 or not isinstance(self._position, int):
            self._errors.append(OrderingMixinException('Position must be a positive number.'))

        if self._position > len(self._destination_objects):
            self._errors.append(OrderingMixinException('Position must be less than the number of destination objects.'))



from __future__ import annotations

from typing import Literal

from django_spire.auth.group.models import AuthGroup
from django_spire.auth.user.models import AuthUser


PERMISSIONS_LEVEL_CHOICES = (
    (0, 'None'),
    (1, 'View'),
    (2, 'Add'),
    (3, 'Change'),
    (4, 'Delete'),
)


VALID_PERMISSION_LEVELS = Literal[0, 1, 2, 3, 4]


# PERMISSION_MODELS_DICT = {
#     'group': {
#         'model': AuthGroup,
#         'is_proxy_model': True
#     },
#     'user': {
#         'model': AuthUser,
#         'is_proxy_model': True
#     }
# }



from __future__ import annotations

from django_spire.ai.chat.tests.test_urls.factories import create_test_chat
from django_spire.core.tests.test_cases import BaseTestCase
from django.urls import reverse


class ChatJsonUrlTests(BaseTestCase):
    def setUp(self) -> None:
        super().setUp()

        self.test_chat = create_test_chat(user=self.super_user)

    def test_delete_view_url_path(self) -> None:
        response = self.client.post(
            reverse(
                'django_spire:ai:chat:json:delete',
                kwargs={'pk': self.test_chat.pk}
            ),
            content_type='application/json'
        )

        assert response.status_code == 200



from __future__ import annotations

from typing import TYPE_CHECKING

from django_spire.ai.chat.router import BaseChatRouter
from django_spire.knowledge.intelligence.workflows.knowledge_workflow import knowledge_search_workflow

if TYPE_CHECKING:
    from dandy.llm.request.message import MessageHistory
    from django.core.handlers.wsgi import WSGIRequest

    from django_spire.ai.chat.message_intel import BaseMessageIntel


class KnowledgeSearchRouter(BaseChatRouter):
    def workflow(
        self,
        request: WSGIRequest,
        user_input: str,
        message_history: MessageHistory | None = None
    ) -> BaseMessageIntel:
        return knowledge_search_workflow(
            request=request,
            user_input=user_input,
            message_history=message_history
        )



from __future__ import annotations

import json

from typing import TYPE_CHECKING, TypeVar

from django.contrib.contenttypes.models import ContentType
from django.db.models import Model

if TYPE_CHECKING:
    from typing import Any

    from django.db.models import QuerySet
    from django.http import HttpRequest


T = TypeVar('T', bound=Model)


def get_object_or_null_obj(queryset_or_model: QuerySet[T] | type[T], **kwargs) -> T:
    """
    Retrieves an object from the given QuerySet or model using the
    provided keyword arguments.

    If the object does not exist, returns a new instance of the model.

    :param queryset_or_model: A QuerySet or model class to query from.
    :param kwargs: Keyword arguments for filtering the object.
    :return: An instance of the model, either retrieved or a new instance if not found.
    """

    if not hasattr(queryset_or_model, 'get'):
        queryset = queryset_or_model._default_manager.all()
    else:
        queryset = queryset_or_model

    try:
        return queryset.get(**kwargs)
    except queryset.model.DoesNotExist:
        return queryset.model()


def get_object_or_none(model: type[T], pk: int, **kwargs) -> Model:
    """
    Retrieves an object from the given model using its primary key and additional filters.

    :param model: The model class to query.
    :param pk: The primary key of the object.
    :param kwargs: Additional filter parameters.
    :return: The model instance if found, otherwise None.
    """

    try:
        return model.objects.get(pk=pk, **kwargs)
    except model.DoesNotExist:
        return None


def process_request_body(request: HttpRequest, key='data') -> Any:
    """
    Processes the HTTP request body and returns the 'data' field from the parsed JSON.

    :param request: The HTTP request object.
    :return: The 'data' field extracted from the JSON body.
    """

    body_unicode = request.body.decode('utf-8')

    if key:
        return json.loads(body_unicode)[key]
    else:
        return json.loads(body_unicode)


def model_object_from_app_label(
    app_label: str,
    model_name: str,
    object_pk: int
) -> Model | None:
    """
    Retrieves a model instance based on the application label, model name, and primary key.

    :param app_label: The application label of the model.
    :param model_name: The name of the model.
    :param object_pk: The primary key of the object.
    :return: The model instance if found, otherwise None.
    """

    try:
        content_type = ContentType.objects.get(
            app_label=app_label,
            model=model_name
        )
    except ContentType.DoesNotExist:
        return None

    model_class = content_type.model_class()
    return get_object_or_none(model_class, pk=object_pk)



from __future__ import annotations

from typing import TYPE_CHECKING

from dandy import Prompt

from django_spire.core.tag.intelligence.tag_set_bot import TagSetBot
from django_spire.core.tag.service.tag_service import BaseTagService
from django_spire.core.tag.tools import simplify_tag_set, simplify_and_weight_tag_set_to_dict, \
    get_score_percentage_from_tag_set_weighted

if TYPE_CHECKING:
    from django_spire.knowledge.collection.models import Collection


class CollectionTagService(BaseTagService['Collection']):
    obj: Collection

    def process_and_set_tags(self):
        collection_prompt = Prompt()

        collection_prompt.sub_heading(self.obj.name)
        collection_prompt.text(self.obj.description)

        tag_set = TagSetBot().process(
            content=collection_prompt
        )

        self.set_tags_from_tag_set(
            tag_set=tag_set,
        )

    def get_aggregated_tag_set(self) -> set[str]:
        tag_set = self.obj.tag_set

        for collection in self.obj.children.active():
            tag_set.update(collection.services.tag.get_aggregated_tag_set())

        for entry in self.obj.entries.active():
            tag_set.update(entry.tag_set)

        return tag_set

    def get_score_percentage_from_aggregated_tag_set_weighted(self, tag_set: set[str]) -> float:
        return get_score_percentage_from_tag_set_weighted(
            tag_set_actual=tag_set,
            tag_set_reference=self.get_aggregated_tag_set()
        )

    def get_simplified_aggregated_tag_set(self) -> set[str]:
        return simplify_tag_set(self.get_aggregated_tag_set())

    def get_simplified_and_weighted_aggregated_tag_set(self) -> dict[str, int]:
        return simplify_and_weight_tag_set_to_dict(self.get_aggregated_tag_set())



from __future__ import annotations

import re

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Locator, Page


class LazyTab:
    """Playwright component for django_spire/lazy_tab/lazy_tab.html"""

    selected_class: str = 'tab-item'

    def __init__(self, page: Page, container_selector: str, tab_id: str | None = None) -> None:
        self.container_selector = container_selector
        self.page = page
        self.tab_id = tab_id

    @property
    def container(self) -> Locator:
        return self.page.locator(self.container_selector)

    @property
    def sections(self) -> Locator:
        return self.container.locator('[role="tabpanel"]')

    @property
    def triggers(self) -> Locator:
        return self.container.locator('[role="tab"]')

    def click_tab(self, index: int) -> None:
        self.get_trigger(index).click()

    def get_section(self, index: int) -> Locator:
        return self.sections.nth(index)

    def get_section_count(self) -> int:
        return self.sections.count()

    def get_trigger(self, index: int) -> Locator:
        return self.triggers.nth(index)

    def get_trigger_count(self) -> int:
        return self.triggers.count()

    def get_url_param(self) -> str | None:
        if not self.tab_id:
            return None

        url = self.page.url
        match = re.search(f'{self.tab_id}=([^&]+)', url)

        if match:
            return match.group(1)

        return None

    def get_visible_section(self) -> Locator | None:
        for i in range(self.sections.count()):
            section = self.sections.nth(i)

            if section.is_visible():
                return section

        return None

    def get_visible_section_text(self) -> str:
        section = self.get_visible_section()

        if section:
            return section.inner_text()

        return ''

    def is_loading(self, index: int) -> bool:
        section = self.get_section(index)
        spinner = section.locator('.spinner-border')
        return spinner.is_visible()

    def is_tab_selected(self, index: int) -> bool:
        trigger = self.get_trigger(index)
        classes = trigger.get_attribute('class') or ''
        return self.selected_class in classes

    def wait_for_section_content(self, index: int, timeout: int = 5000) -> None:
        section = self.get_section(index)
        section.locator('.spinner-border').wait_for(state='hidden', timeout=timeout)

    def wait_for_tabs_to_load(self) -> None:
        self.triggers.first.wait_for()



# Generated by Django 5.2 on 2025-04-23 17:23

import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Viewed',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField(editable=False)),
                ('created_datetime', models.DateTimeField(default=django.utils.timezone.localtime, editable=False)),
                ('content_type', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype')),
                ('user', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='views', related_query_name='view', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'Viewed',
                'verbose_name_plural': 'Views',
                'db_table': 'django_spire_history_viewed',
            },
        ),
    ]



from __future__ import annotations

from django.urls import path

from django_spire.ai.sms import views


app_name = 'sms'

urlpatterns = [
    path('webhook/', views.webhook_view, name='webhook'),
]



from __future__ import annotations

from django_spire.contrib.service import BaseDjangoModelService

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from django_spire.knowledge.collection.models import Collection


class CollectionToolService(BaseDjangoModelService['Collection']):
    obj: Collection

    def get_children_ids(self, parent_id: int) -> set[int]:
        descendant_ids = set()
        current_level_ids = [parent_id]

        while current_level_ids:
            child_ids = self.obj_class.objects.by_parent_ids(
                parent_ids=current_level_ids
            ).values_list('id', flat=True)

            if not child_ids:
                break

            descendant_ids.update(child_ids)
            current_level_ids = child_ids

        return descendant_ids

    def get_root_collection_pk(self) -> int:
        collection = self.obj

        while collection.parent_id:
            collection = collection.parent

        return collection.id



from enum import Enum


def django_choices_to_enums(
        class_name: str,
        choices: list[tuple[str, str]]
):
    choices_dict = {str(k).upper(): k for k, _ in choices}
    return Enum(class_name, choices_dict)



from __future__ import annotations

from enum import StrEnum


class ResponsiveMode(StrEnum):
    """
    It defines how tables handle responsive behavior at different viewport sizes.

    Collapse: Any columns with breakpoint attributes are hidden at smaller viewports
        and shown at their specified breakpoint.

    Scroll: All columns remain visible regardless of viewport size. The table
        container enables horizontal scrolling to accommodate the full width.
    """

    COLLAPSE = 'collapse'
    SCROLL = 'scroll'



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from playwright.sync_api import Locator, Page


class Breadcrumb:
    """Playwright component for django_spire/element/breadcrumb_element.html"""

    def __init__(self, page: Page) -> None:
        self.page = page

    @property
    def breadcrumb(self) -> Locator:
        return self.page.locator('ol.breadcrumb')

    @property
    def items(self) -> Locator:
        return self.breadcrumb.locator('.breadcrumb-item')

    def click_item(self, index: int) -> None:
        link = self.get_item(index).locator('a')

        if link.count() > 0:
            link.click()

    def get_item(self, index: int) -> Locator:
        return self.items.nth(index)

    def get_item_count(self) -> int:
        return self.items.count()

    def get_item_href(self, index: int) -> str | None:
        link = self.get_item(index).locator('a')

        if link.count() > 0:
            return link.get_attribute('href')

        return None

    def get_item_text(self, index: int) -> str:
        return self.get_item(index).inner_text()

    def get_items_text(self) -> list[str]:
        return [self.get_item_text(i) for i in range(self.get_item_count())]

    def get_last_item_text(self) -> str:
        return self.items.last.inner_text()

    def is_item_clickable(self, index: int) -> bool:
        return self.get_item(index).locator('a').count() > 0

    def is_visible(self) -> bool:
        return self.breadcrumb.is_visible()



from __future__ import annotations

from django.db import models


class NotificationPriorityChoices(models.TextChoices):
    LOW = 'low'
    MEDIUM = 'medium'
    HIGH = 'high'


class NotificationStatusChoices(models.TextChoices):
    PENDING = 'pending'
    PROCESSING = 'processing'
    SENT = 'sent'
    ERRORED = 'errored'
    FAILED = 'failed'


class NotificationTypeChoices(models.TextChoices):
    APP = 'app'
    EMAIL = 'email'
    PUSH = 'push'
    SMS = 'sms'
